"""Life Tables - Loads and calculates life tables

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from typing import Dict
from actuarialmath.reserves import Reserves

class LifeTable(Reserves):
    """Calculate life table, and iteratively fill in missing values

    Args:
      udd : assume UDD or constant force of mortality for fractional ages
      verbose : whether to echo update steps

    Notes:
      4 types of information can be loaded and calculated in the life table:

      - 'q' : probability (x) dies in one year
      - 'l' : number of lives aged x
      - 'd' : number of deaths of age x
      - 'p' : probability (x) survives at least one year
    """

    def __init__(self, udd: bool = True, verbose: bool = False, **kwargs):
        super().__init__(udd=udd, **kwargs)
        self._verbose = verbose
        self._table = {'l':{}, 'd':{}, 'q':{}, 'p':{}}  # columns in life table

        # Set basic survival functions by interpolating lifetable integer ages
        def _mu(x: int, s: float) -> float:
            u = math.floor(s)
            return self.mu_r(x, s=u, r=s-u)
        
        def _l(x: int, s: float) -> float:
            u = math.floor(s)
            return self.l_r(x, s=u, r=s-u)
        
        def _S(x: int, s, t: float) -> float:
            u = math.floor(t)   # u+r_t_x = u_t_x * 
            return self.p_x(x, s=s, t=u) * self.p_r(x, s=s+u, t=t-u)
        
        def _f(x: int, s, t: float) -> float:
            u = math.floor(t)   # f_x(u+r) = u_p_x * f_u(r)
            return self.p_x(x, s=s, t=u) * self.f_r(x, s=s+u, t=t-u)

        self.set_survival(mu=_mu, S=_S, f=_f, l=_l, minage=-1, maxage=-1)


    def set_table(self, fill: bool = True, minage: int = -1, maxage: int = -1,
                  l: Dict[int, float] | None = None,
                  d: Dict[int, float] | None = None,
                  p: Dict[int, float] | None = None, 
                  q: Dict[int, float] | None = None) -> "LifeTable":
        """Update life table

        Args:
          l : lives at start of year x, or
          d : deaths in year x, or
          p : probabilities that (x) survives one year, or
          q : probabilities that (x) dies in one year
          fill : whether to automatically fill table cells (default is True)
          minage : minimum age in table
          maxage : maximum age in table
        """

        inputs = {k:v for k,v in {'l':l, 'd':d, 'q':q, 'p':p}.items() if v}

        # infer min and max ages from inputs
        if minage < 0:
            minage = min([min(v) for v in inputs.values()])
            if self._MINAGE < 0 or minage < self._MINAGE:
                self._MINAGE = minage
        else:
            self._MINAGE = minage
        if maxage < 0:
            maxage = max([max(v) for v in inputs.values()])
            if self._MAXAGE < 0:
                self._MAXAGE = maxage + 1
            if maxage > self._MAXAGE:
                self._MAXAGE = maxage
        else:
            self._MAXAGE = maxage

        # update table from inputs
        for label, col in inputs.items():
            self._table[label].update(col)

        # derive and fill table values
        if fill:
            self.fill_table()
        return self

    def fill_table(self, radix: int = Reserves._RADIX) -> "LifeTable":
        """Iteratively fill in missing table cells (does not check consistency)

        Args:
          radix : initial number of lives
        """

        def q_x(x: int) -> float | None:
            """Helper to try compute one-year mortality rate for (x): 1_q_x"""
            if x in self._table['q']:
                return self._table['q'][x]
            if x in self._table['p']:
                return 1 - self._table['p'][x]
            if x in self._table['d'] and x in self._table['l']:
                return self._table['d'][x] / self._table['l'][x]
            return None

        def p_x(x: int) -> float | None:
            """Helper to try compute one-year survival for (x): 1_q_x"""
            if x in self._table['p']:
                return self._table['p'][x]
            if x in self._table['q']:
                return 1 - self._table['q'][x]
            return None

        def l_x(x: int) -> float | None:
            """Helper to try compute number of lives aged x: l_x"""
            if x in self._table['l']:
                return self._table['l'][x]
            if x+1 in self._table['l'] and x in self._table['q']:
                return self._table['l'][x+1] / (1 - self._table['q'][x])
            if x-1 in self._table['l'] and x-1 in self._table['q']:
                return self._table['l'][x-1] * (1 - self._table['q'][x-1])
            if x in self._table['d'] and x in self._table['q']:
                return self._table['d'][x] / self._table['q'][x]
            return None

        def d_x(x: int) -> float | None:
            """Helper to try compute number of deaths in one year for (x): d_x"""
            if x in self._table['d']:
                return self._table['d'][x]
            if x+1 in self._table['l'] and x in self._table['l']:
                return self._table['l'][x] - self._table['l'][x+1]
            else:
                return None

        # Iterate a few times to impute life table values
        funs = {'l': l_x, 'd': d_x, 'q': q_x, 'p': p_x}
        updated = 0
        for loop in range(2):   # loop second time if radix needed
            prev = updated - 1
            while updated != prev:        # continue while changes  
                prev = updated
                for col, fun in funs.items():  # loop columns
                    for x in range(self._MINAGE, self._MAXAGE + 1):  # loop ages
                        if x not in self._table[col]:
                            value = fun(x)
                            if value is not None:   # update value 
                                self._table[col][x] = round(value, 7)
                                updated += 1        # increment counter of changes
                                if self._verbose:
                                    print(f"{updated} {col}(x={x}) = {value}")
            if not self._table['l']:  # assume starting number of lives if necc
                self._table['l'][self._MINAGE] = radix
        return self

    def mu_x(self, x: int, s: int = 0, t: int = 0) -> float:
        """Compute mu_x from p_x in life table

        Args:
          x : age of selection
          s : years after selection
          t : death within next t years
        """
        return -math.log(max(0.00001, self.p_x(x, s=s+t, t=1)))

    def l_x(self, x: int, s: int = 0) -> float:
        """Lookup l_x from life table

        Args:
          x : age of selection
          s : years after selection
        """
        if x+s in self._table['l']:
            return self._table['l'][x+s]
        else:
            return 0

    def d_x(self, x: int, s: int = 0, t: int = 1) -> float:
        """Compute deaths as lives at x_t divided by lives at x

        Args:
          x : age of selection
          s : years after selection
          t : death within next t years
        """
        if x+s+t <= self._MAXAGE:
            return self.l_x(x, s=s) - self.l_x(x, s=s+t)
        else:
            return 0.

    def p_x(self, x: int, s: int = 0, t: int = 1) -> float:
        """t_p_x = lives beginning year x+t divided lives beginning year x

        Args:
          x : age of selection
          s : years after selection
          t : death within next t years
        """
        denom = self.l_x(x, s=s)
        if denom and x+s+t <= self._MAXAGE:
            return self.l_x(x, s=s+t) / denom
        else:
            return 0.   # in the long term, we are all dead

    def q_x(self, x: int, s: int = 0, t: int = 1, u: int = 0) -> float:
        """Deferred mortality: u|t_q_x = (l[x+u] - l[x+u+t]) / l[x]

        Args:
          x : age of selection
          s : years after selection
          u : survive u years, then...
          t : death within next t years
        """
        denom = self.l_x(x, s=s)
        if denom and x+s+t <= self._MAXAGE:
            return self.d_x(x, s=s+u, t=t) / self.l_x(x, s=s)
        else:
            return 1     # the only certainty in life

    def e_x(self, x: int, s: int = 0, n: int = Reserves.WHOLE) -> float:
        """Expected curtate lifetime from sum of lives in table

        Args:
          x : age of selection
          s : years after selection
          n : future lifetime limited at n years
        """
        # E[K_x] = sum([self.p(x, k+1) for k in range(n)])
        n = min(self._MAXAGE - x, n)
        e = sum([self.l(x+1+s) for s in range(n)]) # since s_p_x = l_x+s / l_x
        return e

    def E_x(self, x: int, s: int = 0, t: int = 1, moment: int = 1) -> float:
        """Pure Endowment from life table and interest rate

        Args:
          x : age of selection
          s : years after selection
          t : survives t years
          moment : return first (1) or second (2) moment or variance (-2)
        """
        if t == 0:
            return 1.
        if t < 0:
            return 0.
        t = self.max_term(x+s, t)
        p = self.l_x(x, s=s+t) / self.l_x(x, s=s)
        if moment == self.VARIANCE:
            return self.interest.v_t(t)**moment * p * (1 - p)
        if moment == 1:
            return self.interest.v_t(t) * p
        # SULT shortcut: t_E_x(moment=2) = t_E_x(moment=1) * v**t
        return self.interest.v_t(t)**(moment-1) * self.E_x(x, s=s, t=t)

    def __getitem__(self, col: str) -> Dict[int, float]:
        """Returns a column of the life table

        Args:
          col : name of life table column to return
        """
        assert col[0] in self._table, f"must be one of {list(self._table.keys())}"
        return self._table[col[0]]

    def frame(self) -> pd.DataFrame:
        """Return life table columns and values in a DataFrame"""
        return pd.DataFrame.from_dict(self._table).sort_index(axis=0)


if __name__ == "__main__":
    print("SOA Question 6.53:  (D) 720")
    x = 0
    life = LifeTable().set_interest(i=0.08)\
                      .set_table(q={x: 0.1, x+1: 0.1, x+2: 0.1})
    A = life.term_insurance(x, t=3)
    G = life.gross_premium(a=1, A=A, benefit=2000, initial_premium=0.35)
    print(A, G)
    print(life.frame())
    print()

    print("SOA Question 6.41:  (B) 1417")
    x = 0
    life = LifeTable().set_interest(i=0.05)\
                      .set_table(q={x:.01, x+1:.02})
    P = 1416.93
    a = 1 + life.E_x(x, t=1) * 1.01
    A = (life.deferred_insurance(x, u=0, t=1) 
         + 1.01 * life.deferred_insurance(x, u=1, t=1))
    print(a, A)
    P = 100000 * A / a
    print(P)
    print(life.frame())
    print()
    

    print("SOA Question 3.11:  (B) 0.03")
    life = LifeTable(udd=True).set_table(q={50//2: .02, 52//2: .04})
    print(life.q_r(50//2, t=2.5/2))
    print(life.frame())
    print()
    

    print("SOA Question 3.5:  (E) 106")
    l = {60 + x: l * 11111 for x,l in enumerate([9, 8, 7, 6, 5, 4, 3, 2])}
    a, b = (LifeTable(udd=udd).set_table(l=l).q_r(60, u=3.4, t=2.5)
            for udd in [True, False])
    print(100000 * (a - b))
    print()
    

    print("SOA Question 3.14:  (C) 0.345")
    life = LifeTable(udd=True).set_table(l={90: 1000, 93: 825},
                                         d={97: 72},
                                         p={96: .2},
                                         q={95: .4, 97: 1})
    print(life.q_r(90, u=93-90, t=95.5-93))
    print(life.frame())
    print()



"""Future lifetimes - Computes expectations and moments of future lifetime

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
import math
from actuarialmath.survival import Survival

class Lifetime(Survival):
    """Computes expected moments of future lifetime"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


    def e_x(self, x: int, s: int = 0, t: int = Survival.WHOLE, 
            curtate: bool = True, moment: int = 1) -> float:
        """Compute curtate or complete expectations and moments of life

        Args:
          x : age of selection
          s : years after selection
          t : limited at t years
          curtate : whether curtate (True) or complete (False) expectations
          moment : whether to compute first (1) or second (2) moment

        Examples:
          >>> def l(x, s): return 0. if (x+s) >= 100 else 1 - ((x + s)**2) / 10000.
          >>> print(Lifetime().set_survival(l=l).e_x(75, t=10, curtate=False))
        """
        assert moment in [1, 2, self.VARIANCE], "moment must be 1, 2 or -2"
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"

        if t == 1 and curtate:      # shortcut for e_x:1
            return self.p_x(x, s=s, t=1)
        
        t = self.max_term(x+s, t)   # length of term must be bounded by max age
        if curtate:
            if moment == 1:
                return sum([self.p_x(x, s=s, t=k)
                            for k in range(1, round(t+1))]) 
            e2 = sum([((2 * k) - 1) * self.p_x(x, s=s, t=k)
                      for k in range(1, round(t+1))])
        else:
            if moment == 1:
                return self.integral(lambda t: self.S(x, s, t), 0., float(t))
            e2 = self.integral(lambda t: 2 * t * self.S(x, s, t), 0., float(t))

        if moment == self.VARIANCE:  # variance is E[T_x^2] - E[T_x]^2
            return e2 - self.e(x, s=s, t=t, curtate=curtate, moment=1)**2
        return e2   # return second moment


if __name__ == "__main__":
    print("SOA Question 2.4: (E) 8.2")
    def l(x, s): return 0. if (x+s) >= 100 else 1 - ((x + s)**2) / 10000.
    e = Lifetime().set_survival(l=l).e_x(75, t=10, curtate=False)
    print(e)

    print("SOA Question 2.1: (B) 2.5")
    def fun(omega):  # Solve first for omega, given mu_65 = 1/180
        return Lifetime().set_survival(l=lambda x,s: (1-(x+s)/omega)**0.25,
                                       maxage=omega).mu_x(65)
    omega = int(Lifetime.solve(fun, target=1/180, grid=100))  # solve for omega
    e = Lifetime().set_survival(l=lambda x,s: (1 - (x+s)/omega)**0.25, 
                                maxage=omega).e_x(106)
    print(e)    
    



"""Fractional age assumptions- Computes survival and mortality functions between integer ages.

MIT License. Copyright (c) 2022-2023 Terence Lim
"""

import math
from actuarialmath.lifetime import Lifetime

class Fractional(Lifetime):
    """Compute survival functions at fractional ages and durations

    Args:
      udd : select UDD (True, default) or CFM (False) between integer ages
    """

    def __init__(self, udd: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.udd_ = udd

    #
    # Define actuarial forms of survival functions at fractional ages
    # 
    def l_r(self, x: int, s: int = 0, r: float = 0.) -> float:
        """Number of lives at fractional age: l_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        s += math.floor(r)  # interpolate lives between consecutive integer ages
        r -= math.floor(r)
        if r == 0:
            return self.l_x(x, s=s)
        if r == 1.0:
            return self.l_x(x, s=s+1)
        if self.udd_:
            return self.l_x(x, s=s)*(1-r) + self.l_x(x, s=s+1)*r
        else:
            return self.l_x(x, s=s)**(1-r) * self.l_x(x, s=s+1)**r

    def p_r(self, x: int, s: int = 0, r: float = 0., t: float = 1.) -> float:
        """Probability of survival from and through fractional age: t_p_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : fractional number of years survived

        Examples:
        >>> life = Fractional(udd=False).set_survival(l=lambda x,t: 50-x-t)
        >>> print(life.p_r(47, r=0.), life.p_r(47, r=0.5), life.p_r(47, r=1.))

        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        assert t >= 0, "t must be non-negative"
        r_floor = math.floor(r)
        s += r_floor
        r -= r_floor
        if 0. <= r + t <= 1.:
            if self.udd_:
                return 1 - self.q_r(x, s=s, r=r, t=t)
            else:          # Constant force shortcut within int age
                return self.p_x(x, s=s)**t   # does not depend on r
        return self.l_r(x, s=s, r=r+t) / self.l_r(x, s=s, r=r)

    def q_r(self, x: int, s: int = 0, r: float = 0., t: float = 1., 
            u: float = 0.) -> float:
        """Deferred mortality rate within fractional ages: u|t_q_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          u : fractional number of years survived, then
          t : death within next fractional years t

        Examples:
          >>> life = Fractional(udd=False).set_survival(l=lambda x,t: 50-x-t)
          >>> print(life.q_r(x, r=0.5))

        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        assert t >= 0, "t must be non-negative"
        assert t >= 0, "t must be non-negative"
        r_floor = math.floor(r)
        s += r_floor
        r -= r_floor
        if 0 <= r + t + u <= 1:
            if u > 0:      # Die within u|t_q == Die in t+u but not in u 
                return self.q_r(x, s=s, r=r, t=u+t) - self.q_r(x, s=s, r=r, t=u)
            if self.udd_:  # UDD shortcut within integer age
                return (t * self.q_x(x, s=s)) / (1. - r * self.q_x(x, s=s))
            else:
                return 1 - self.p_r(x, s=s, r=r, t=t)
        return ((self.l_r(x, s=s, r=r+u) - self.l_r(x, s=s, r=r+u+t)) 
                / self.l_r(x, s=s, r=r))
        
    def mu_r(self, x: int, s: int = 0, r: float = 0.) -> float:
        """Force of mortality at fractional age: mu_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        r_floor = math.floor(r)
        s += r_floor
        r -= r_floor
        if self.udd_:   # UDD shortcut
            return self.q_x(x, s=s) / (1. - r*self.q_x(x, s=s))
        else:          # Constant force shortcut
            return -math.log(max(0.0001, self.p_x(x, s=s)))

    def f_r(self, x: int, s: int = 0, r: float = 0., t: float = 0.0) -> float:
        """mortality function at fractional age: f_[x]+s+r (t)

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : death at fractional year t
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        assert t >= 0, "t must be non-negative"
        if 0. <= r + t <= 1.:   # shortcuts available within integer ages
            if self.udd_:           # UDD constant q_x
                return self.q_x(x, s=s)      # does not depent on fractional age
            else:                  # Constant force shortcut
                mu = -math.log(max(0.00001, self.p_x(x, s=s)))
                return math.exp(-mu*t) * mu  # does not depend on fractional age
        else:      # survive to integer age then extend by fractional mortality
            r_floor = math.floor(r)
            r -= r_floor            # s.t. r < 1 
            s += r_floor            # while maintaining x+s+r unchanged
            t_floor = math.floor(r + t)
            u = t_floor - r         # s.t. u + r is integer
            t = t + r - t_floor     # s.t. t < 1
            return self.p_r(x, s=s, r=r, t=u) * self.f_r(x, s=s+u+r, t=t)

    #
    # Define fractional age pure endowment function
    #
        
    def E_r(self, x: int, s: int = 0, r: float = 0., t: float = 1.) -> float:
        """Pure endowment at fractional age: t_E_[x]+s+r

        Args:
          x : age of selection
          s : years after selection
          r : fractional year after selection
          t : limited at fractional year t
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        assert r >= 0, "r must be non-negative"
        assert t >= 0, "t must be non-negative"
        return self.p_r(x, s=s, r=r, t=t) * self.interest.v_t(t)

    #
    # Define fractional age expectations of future lifetime
    #
    def e_r(self, x: int, s: int = 0, t: float = Lifetime.WHOLE) -> float:
        """Temporary expected future lifetime at fractional age: e_[x]+s:t

        Args:
          x : age of selection
          s : years after selection
          t : fractional year limit of expected future lifetime
        """ 
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        if t == 0:
            return 0
        elif t < 0:   # shortcuts for complete expectation 
            if self.udd_:  # UDD case
                return self.e_x(x, s=s, t=t, curtate=True) + 0.5
            else:         # Constant Force Case: compute as maxage temporary
                return self.e_r(x, s=s, t=self.max_term(x+s, t))
        elif t <= 1:  # shortcuts within integer age
            if self.udd_:  # UDD case
                if t == 1:   # shortcut for UDD 1-year temporary expectation
                    return 1. - self.q_x(x, s=s)*(1/2)
                else:        # UDD formula for fractional temporary expectation
                    return (self.q_r(x, s=s, t=t) * (t/2)
                            + self.p_r(x, s=s, t=t) * t)
            else:         # Constant Force case
                mu = -math.log(max(0.00001, self.p_x(x, s=s)))  # constant mu
                return (1. - math.exp(-mu*t)) / mu   # constant force formula
        else:  # apply one-year recursion formula
            return (self.e_x(x, s=s, t=1) +
                    (self.p_x(x, s=s, t=1) * self.e_r(x, s=s+1, t=t-1)))

    #
    # Approximation of curtate and complete lifetimes
    #
    @staticmethod 
    def e_approximate(e_complete: float = None, e_curtate: float = None) -> float:
        """Convert between curtate and complete expectations assuming UDD shortcut

        Args:
          e_complete : complete expected lifetime
          e_curtate : or curtate expected lifetime

        Returns:
          approximate complete or curtate expectation assuming UDD
        
        Examples:
          >>> print(Fractional.e_curtate(e_complete=15))
          >>> print(Fractional.e_curtate(e_curtate=15))
        """
        if e_complete is not None:
            assert e_curtate is None, "one of e and e_curtate must be None"
            return e_complete - 0.5
        else:
            return e_curtate + 0.5

if __name__ == "__main__":
    print(Fractional.e_approximate(e_complete=15))  # output e_curtate
    print(Fractional.e_approximate(e_curtate=15))   # output e_complete

    x = 45
    life = Fractional(udd=False).set_survival(l=lambda x,t: 50-x-t)
    print(life.q_r(x, r=0.), life.q_r(x, r=0.5), life.q_r(x, r=1.))
    life = Fractional(udd=True).set_survival(l=lambda x,t: 50-x-t)
    print(life.q_r(x, r=0.), life.q_r(x, r=0.5), life.q_r(x, r=1.))



"""Survival models - Computes survival and mortality functions

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
from typing import Callable, Tuple, Any, List
import math
import numpy as np
from actuarialmath.life import Life

class Survival(Life):
    """Set and derive basic survival and mortality functions"""
    _RADIX = 100000   # default initial number of lives in life table

    def set_survival(self,
                     S: Callable[[int,float,float], float] | None = None, 
                     f: Callable[[int,float,float], float] | None = None,
                     l: Callable[[int,float], float] | None = None, 
                     mu: Callable[[int,float], float] | None = None,
                     maxage: int = Life._MAXAGE,
                     minage: int = Life._MINAGE) -> "Survival":
        """Construct the basic survival and mortality functions given any one form

        Args:
          S : probability [x]+s survives t years
          f : or mortality of [x]+s after t years 
          l : or number of lives aged (x+t)
          mu : or force of mortality at age (x+t)
          maxage : maximum age
          minage : minimum age
        """
        self._MAXAGE = maxage
        self._MINAGE = minage
        self.S = None   # survival probability: Prob(T_[x]+s > t)
        self.f = None   # lifetime density: f_[x]+s(t) ~ Prob[([x]+s) dies at t]
        self.l = None   # number of lives aged [x]+s: l_[x]+s
        self.mu = None  # force of mortality: mu_(x+t)

        def S_from_l(x: int, s, t: float) -> float:
            """Derive survival probability from number of lives"""
            return (self.l(x, s+t) / self.l(x, s)) if self.l(x, s) else 0.

        def mu_from_l(x: int, t: float) -> float:
            """Derive force of mortality from number of lives"""
            return -self.derivative(lambda s: math.log(self.l(x, s)), t)

        def f_from_l(x: int, s, t: float) -> float:
            """Derive lifetime density function from number of lives"""
            return -self.derivative(lambda t: self.l(x, s+t), t)

        def mu_from_S(x: int, t: float) -> float:
            """Derive force of mortality from survival probability"""
            return -self.derivative(lambda s: math.log(self.S(x, 0, s)), t)

        def f_from_S(x: int, s, t: float) -> float:
            """Derive lifetime density function from survival probability"""
            return -self.derivative(lambda t: self.S(x, s, t), t)

        def S_from_mu(x: int, s, t: float) -> float:
            """Derive survival probability from force of mortality"""
            return math.exp(-self.integral(lambda t: self.mu(x, s+t), 0, t))

        def S_from_f(x: int, s, t: float) -> float:
            """Derive survival probability from lifetime density function"""
            return 1 - self.integral(lambda t: self.f(x, s, t), 0, t)

        def f_from_mu(x: int, s, t: float) -> float:
            """Derive lifetime density function from force of mortality"""
            return self.S(x, s, t) * self.mu(x, s+t)

        def mu_from_f(x: int, t: float) -> float:
            """Derive force of mortality from lifetime density function"""
            self.mu = self.f(x, 0, t) / self.S(x, 0, t)

        # derive and set all forms of basic survival and mortality functions
        if l is not None:
            assert callable(l), "l must be callable"
            self.S = S_from_l
            self.mu = mu_from_l
            self.f = f_from_l
        if S is not None:
            assert callable(S), "S must be callable"
            self.mu = mu_from_S
            self.f = f_from_S
        if mu is not None:
            assert callable(mu), "mu must be callable"
            self.S = S_from_mu
            self.f = f_from_mu
        if f is not None:
            assert callable(f), "f must be callable"
            self.S = S_from_f
            self.mu = mu_from_f
        self.l = l or self.l
        self.S = S or self.S
        self.f = f or self.f
        self.mu = mu or self.mu
        return self


    #
    #  Actuarial forms of survival and mortality functions, at integer ages
    #
    def l_x(self, x: int, s: int = 0) -> float:
        """Number of lives at integer age [x]+s: l_[x]+s

        Args:
          x : age of selection
          s : years after selection
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        if self.l is not None:
            return self.l(x, s)
        return Life.LIFES * self.p_x(x=self._MINAGE, s=0, t=s+x-self._MINAGE)

    def d_x(self, x: int, s: int = 0) -> float:
        """Number of deaths at integer age [x]+s: d_[x]+s

        Args:
          x : age of selection
          s : years after selection
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        return self.l_x(x=x, s=s) - self.l_x(x=x, s=s+1)

    def p_x(self, x: int, s: int = 0, t: int = 1) -> float:
        """Probability that [x]+s lives another t years: : t_p_[x]+s

        Args:
          x : age of selection
          s : years after selection
          t : survives at least t years
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        if self.S is not None:
            return self.S(x, s, t)
        raise Exception("No functions implemented to compute survival")

    def q_x(self, x: int, s: int = 0, t: int = 1, u: int = 0) -> float:
        """Probability that [x]+s lives for u, but not t+u years: u|t_q_[x]+s

        Args:
          x : age of selection
          s : years after selection
          u : survives u years, then
          t : dies within next t years
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        return self.p_x(x, s=s, t=u) - self.p_x(x, s=s, t=t+u)

    def f_x(self, x: int, s: int = 0, t: int = 0) -> float:
        """Lifetime density function of [x]+s after t years: f_[x]+s(t)

        Args:
          x : age of selection
          s : years after selection
          t : dies at year t
        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        if self.f is not None:
            return self.f(x, s, t)
        return self.p_x(x, s=s, t=t) * self.mu_x(x, s=s, t=t)

    def mu_x(self, x: int, s: int = 0, t: int = 0) -> float:
        """Force of mortality of [x] at  s+t years: mu_[x](s+t)

        Args:
          x : age of selection
          s : years after selection
          t : force of mortality at year t

        Examples:
          >>> def S(x, s): return (1 - (x+s) / 60)**(1 / 3)
          >>> print(Survival().set_survival(l=S).mu_x(35))

        """
        assert x >= 0, "x must be non-negative"
        assert s >= 0, "s must be non-negative"
        if self.mu is not None:
            return self.mu(x, s+t)
        return self.f_x(x, s=s, t=t) / self.p_x(x, s=s, t=t)

#    def survival_curve(self, x: int, s: int = 0, stop: int = 0) -> Tuple[List, List]:
#        """Construct curve of survival probabilities at each integer age
#
#        Args:
#          x : age of selection
#          s : years after selection
#          stop : end at time t, inclusive
#
#        Returns:
#          lists of lifetime and survival probability from t, S_x(t) respectively
#        """
#        stop = stop or self._MAXAGE - (x + s)
#        steps = range(stop + 1)
#        return steps, [self.p_x(x=x, s=s, t=t) for t in steps]

if __name__ == "__main__":
    print("SOA Question 2.3: (A) 0.0483")
    B, c = 0.00027, 1.1
    def S(x,s,t): return (math.exp(-B * c**(x+s) * (c**t - 1)/math.log(c)))
    f = Survival().set_survival(S=S).f_x(x=50, t=10)
    print(f)
    
    print("SOA Question 2.6: (C) 13.3")
    def S(x, s): return (1 - (x+s) / 60)**(1 / 3)
    mu = Survival().set_survival(l=S).mu_x(35) * 1000
    print(mu)

    print("SOA Question 2.7: (B) 0.1477")
    def S(x, s): return (1 - ((x+s)/250)) if x+s < 40 else (1 - ((x+s)/100)**2)
    q = Survival().set_survival(l=S).q_x(30, t=20)
    print(q)

    print("CAS41-F99:12: k = 41")
    def mu_from_l(k):
        def S(x, s): return 100*(k - 0.5*(x+s))**(2/3)
        return Survival().set_survival(l=S).mu_x(50)
    k = int(Survival.solve(mu_from_l, target=1/48, grid=50))
    print(k)



"""Premiums - Computes net and gross premiums, with the equivalence principle

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
from actuarialmath.annuity import Annuity

class Premiums(Annuity):
    """Compute et and gross premiums under equivalence principle"""

    #
    # Net level premiums for special insurances
    #
    def net_premium(self, x: int, s: int = 0, t: int = Annuity._WHOLE, 
                    u: int = 0, n: int = 0, b: int = 1,
                    endowment: int = 0, discrete: bool | None = True, 
                    return_premium: bool = False, annuity: bool = False,
                    initial_cost: float = 0.) -> float:
        """Net level premium for special n-pay, u-deferred t-year term insurance

        Args:
          x : age initially insured
          s : years after selection
          u : years of deferral
          n : number of years of premiums paid
          t : year of death
          b : benefit amount
          endowment : endowment amount
          initial_cost : EPV of any other expenses or benefits, if any
          return_premium : if premiums without interest refunded at death 
          annuity : whether benefit is insurance (False) or deferred annuity
          discrete : whether annuity due (True) or continuous (False)
        """
        if annuity:
            A = self.deferred_annuity(x, s=s, b=b, t=t, u=u, 
                                      discrete=discrete or discrete is None)
        else:
            A = self.deferred_insurance(x, s=s, b=b, t=t, u=u, 
                                        discrete=discrete or discrete is None)
            if endowment:
                A += self.E_x(x, s=s, t=t+u) * endowment
        if n == 0:      # if n not specified
            n = u or t  # then set to defer period if any, else same term t
        discrete = discrete or discrete is None  # discrete or semi-discrete
        a = self.temporary_annuity(x, s=s, t=n, discrete=discrete)
        if return_premium: # death benefit include premiums returned w/o interest
            a -= self.increasing_insurance(x, s=s, t=n, discrete=discrete)
        return (A + initial_cost) / a

    #
    # Equivalence principle for WL, Endowment Insurance, and their Annuity twins
    #
    def insurance_equivalence(self, premium: float, b: int = 1,
                              discrete: bool = True) -> float:
        """Compute whole life or endowment insurance factor, given net premium

        Args:
          premium : level net premium amount
          b : benefit amount
          discrete : discrete/annuity due (True) or continuous (False)

        Returns:
          Insurance factor value, given net premium under equivalence principle
        """
        d = self.interest.d if discrete else self.interest.delta
        return premium / (d*b + premium)  # from P = b[dA/(1-A)]

    def annuity_equivalence(self, premium: float, b: int = 1,
                            discrete: bool = True) -> float:
        """Compute whole life or temporary annuity factor, given net premium

        Args:
          premium : level net premium amount
          b : benefit amount
          discrete : discrete/annuity due (True) or continuous (False)

        Returns:
          Annuity factor value, given net premium under equivalence principle
        """
        d = self.interest.d if discrete else self.interest.delta
        return b / (d*b + premium)  # from P = b * (1/a - d)

    def premium_equivalence(self, A: float | None = None,
                            a: float | None = None,
                            b: int = 1, discrete: bool = True) -> float:
        """Compute premium from whole life or endowment insurance and annuity factors

        Args:
          A : insurance factor
          a : annuity factor
          b : insurance benefit amount
          discrete : annuity due (True) or continuous (False)

        Returns:
          Net premium under equivalence principle
        """
        interest = self.interest.d if discrete else self.interest.delta
        if a is None:   # Annuity not given => use shortcut for insurance
            return b * interest * A / (1 - A)
        elif A is None: # Insurance not given => use shortcut for annuity
            return b * (1/a - interest)
        else:           # Both (special) insurance and annuity factors given
            return b * A / a

    #
    # Gross premiums for special insurances
    #
    def gross_premium(self, a: float | None = None, 
                      A: float | None = None, IA: float = 0, 
                      discrete: bool = True, benefit: float = 1,
                      E: float = 0, endowment: int = 0, 
                      settlement_policy: float = 0.,
                      initial_policy: float = 0.,
                      initial_premium: float = 0.,
                      renewal_policy: float = 0., 
                      renewal_premium: float = 0.) -> float:
        """Gross premium by equivalence principle

        Args:
          A : insurance factor
          a : annuity factor
          IA : increasing insurance factor, to return premiums w/o interest
          E : pure endowment factor for endowment benefit
          benefit : insurance benefit amount
          endowment : endowment benefit amount
          settlement_policy : settlement expense per policy
          initial_policy : initial expense per policy
          renewal_policy : renewal expense per policy
          initial_premium : initial premium per $ of gross premium
          renewal_premium : renewal premium per $ of gross premium
          discrete : annuity due (True) or continuous (False)
        """
        if a is None:    # assume WL or Endowment Insurance for twin
            a = self.annuity_twin(A, discrete=discrete)
        elif A is None:  # assume WL or Endowment Insurance for twin
             A = self.insurance_twin(a, discrete=discrete)

        assert endowment == 0 or E > 0  # missing pure endowment factor if needed
        per_premium = renewal_premium * a + (initial_premium - renewal_premium)
        per_policy = renewal_policy * a + (initial_policy - renewal_policy)
        return (((A*(benefit + settlement_policy) + per_policy) + (E*endowment))
                / (a - per_premium - IA))  # IA returns premium w/o interest


if __name__ == "__main__":
    import numpy as np
    
    print("SOA Question 6.29  (B) 20.5")
    life = Premiums().set_interest(i=0.035)
    def fun(a):
        return life.gross_premium(A=life.insurance_twin(a=a),
                                  a=a,
                                  benefit=100000, 
                                  initial_policy=200,
                                  initial_premium=.5,
                                  renewal_policy=50,
                                  renewal_premium=.1)
    print(life.solve(fun, target=1770, grid=[20, 22]))
    print()
    
    print("SOA Question 6.2: (E) 3604")
    life = Premiums()
    A, IA, a = 0.17094, 0.96728, 6.8865
    print(life.gross_premium(a=a,
                             A=A,
                             IA=IA,
                             benefit=100000,
                             initial_premium=0.5,
                             renewal_premium=.05,
                             renewal_policy=200,
                             initial_policy=200))
    print()

    print("SOA Question 6.16: (A) 2408.6")
    life = Premiums().set_interest(d=0.05)
    A = life.insurance_equivalence(premium=2143, b=100000)
    a = life.annuity_equivalence(premium=2143, b=100000)
    p = life.gross_premium(A=A,
                           a=a,
                           benefit=100000,
                           settlement_policy=0,
                           initial_policy=250,
                           initial_premium=.04+.35,
                           renewal_policy=50,
                           renewal_premium=.04+.02) 
    print(A, a, p)
    print()

    print("SOA Question 6.20:  (B) 459")
    l = lambda x,s: dict(zip([75, 76, 77, 78],
                             np.cumprod([1, .9, .88, .85]))).get(x+s, 0)
    life = Premiums().set_interest(i=0.04).set_survival(l=l)
    a = life.temporary_annuity(75, t=3)
    IA = life.increasing_insurance(75, t=2)
    A = life.deferred_insurance(75, u=2, t=1)
    print(life.solve(lambda P: P*IA + A*10000 - P*a, target=0, grid=100))
    print()

    print("Other usage")
    life = Premiums().set_interest(delta=0.06)\
                     .set_survival(mu=lambda x,s: 0.04)
    print(life.net_premium(0))

    print("SOA Question 5.6:  (D) 1200")
    life = Premiums().set_interest(i=0.05)
    var = life.annuity_variance(A2=0.22, A1=0.45)
    mean = life.annuity_twin(A=0.45)
    fund = life.portfolio_percentile(mean, var, prob=.95, N=100)
    print(fund)



"""Life annuities - Computes present values of annuity benefits

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
from typing import Callable, Any
import math
import matplotlib.pyplot as plt
import numpy as np
from actuarialmath.insurance import Insurance

class Annuity(Insurance):
    """Compute present values and relationships of life annuities"""

    def a_x(self, x: int, s: int = 0, t: int = Insurance.WHOLE, u: int = 0,
            benefit: Callable = lambda x,t: 1, discrete: bool = True) -> float:
        """Numerically compute EPV of annuities from survival functions

        Args:
          x : age of selection
          s : years after selection
          u : year deferred
          t : term of insurance
          benefit : benefit as a function of age and year
          discrete : annuity due (True) or continuous (False)
        """
        t = self.max_term(x+s+u, t=t)
        if discrete:
            a = sum([benefit(x+s, k) * self.interest.v_t(k) 
                     * self.p_x(x, s=s, t=k) for k in range(u, t+u)])
        else:   # use continous first principles
            Y = lambda t: (benefit(x+s, t+u) * self.interest.v_t(t+u) 
                           * self.S(x, 0, t=t+u))
            a = self.integral(Y, 0, t)
        return a


    def immediate_annuity(self, x: int, s: int = 0, t: int = Insurance.WHOLE, 
                          b: int = 1, variance=False) -> float:
        """Compute EPV of immediate life annuity

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : benefit amount
          variance : return EPV (False) or variance (True)
        """
        if variance:
            return self.temporary_annuity(x, s=s, t=self.add_term(t, 1), b=b,
                                          discrete=True, variance=True)
        return (self.temporary_annuity(x, s=s, t=t, discrete=True) 
                - 1 + self.E_x(x, s=s, t=t)) * b
    
    def annuity_twin(self, A: float, discrete: bool = True) -> float:
        """Returns annuity from its WL or Endowment Insurance twin"

        Args:
          A : cost of insurance
          discrete : discrete/annuity due (True) or continous (False)
        """
        interest = (self.interest.d if discrete else self.interest.delta)
        return ((1-A) / interest) if interest else 0 # undefined for 0 interest

    def insurance_twin(self, a: float, moment: int = 1, 
                       discrete: bool = True) -> float:
        """Returns WL or Endowment Insurance twin from annuity

        Args:
          a : cost of annuity
          discrete : discrete/annuity due (True) or continous (False)
        """
        assert moment in [1]
        return 1 - a*(self.interest.d if discrete else self.interest.delta)
  
    def annuity_variance(self, A2: float, A1: float, b: float = 1.,
                         discrete: bool = True) -> float:
        """Compute variance from WL or endowment insurance twin

        Args:
          A2 : second moment of insurance factor
          A1 : first moment of insurance factor
          b : annuity benefit amount
          discrete : discrete/annuity due (True) or continous (False)
        """
        return (b**2 * self.insurance_variance(A2=A2, A1=A1)
                / (self.interest.d if discrete else self.interest.delta)**2)
        
    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1, 
                           variance: bool = False, 
                           discrete: bool = True) -> float:
        """Whole life annuity: a_x

        Args:
          x : age of selection
          s : years after selection
          b : annuity benefit amount
          variance : return EPV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        interest = self.interest.d if discrete else self.interest.delta
        if variance:  # short cut for variance of whole life
            A1 = self.whole_life_insurance(x, s=s, moment=1, discrete=discrete)
            A2 = self.whole_life_insurance(x, s=s, moment=2, discrete=discrete)
            return self.annuity_variance(A2=A2, A1=A1, discrete=discrete, b=b)
        A = self.whole_life_insurance(x, s=s, discrete=discrete)
        return b * (1 - A) / interest
            
    def temporary_annuity(self, x: int, s: int = 0, t: int = Insurance.WHOLE, 
                          b: int = 1, variance: bool = False, 
                          discrete: bool = True) -> float:
        """Temporary life annuity: a_x:t

        Args:
          x : age of selection
          s : years after selection
          t : term of annuity in years
          b (int) : annuity benefit amount
          variance : return EPV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.endowment_insurance(x, s=s, t=t, discrete=discrete)
            A2 = self.endowment_insurance(x, s=s, t=t, moment=2, 
                                          discrete=discrete)
            return self.annuity_variance(A2=A2, A1=A1, discrete=discrete, b=b)

        # difference of whole life on (x) and deferred whole life on (x+t)
        a = self.whole_life_annuity(x, s=s, b=b, discrete=discrete)
        if t < 0 or self.max_term(x+s, t) < t:
            return a
        a_t = self.whole_life_annuity(x, s=s+t, b=b, discrete=discrete)
        return a - (a_t * self.E_x(x, s=s, t=t))

    def deferred_annuity(self, x: int, s: int = 0, u: int = 0, 
                         t: int = Insurance.WHOLE, b: int = 1, 
                         discrete: bool = True) -> float:
        """Deferred life annuity n|t_a_x =  n+t_a_x - n_a_x

        Args:
          x : age of selection
          s : years after selection
          u : years deferred
          t : term of annuity in years
          b : annuity benefit amount
          discrete : annuity due (True) or continuous (False)
        """
        a = self.temporary_annuity(x, s=s+u, t=t, b=b, discrete=discrete)
        return self.E_x(x, s=s, t=u) * a

    def certain_life_annuity(self, x: int, s: int = 0, u: int = 0, 
                             t: int = Insurance.WHOLE, b: int = 1, 
                             discrete: bool = True) -> float:
        """Certain and life annuity = certain + deferred

        Args:
          x : age of selection
          s : years after selection
          u : years of certain annuity
          t : term of life annuity in years
          b : annuity benefit amount
          discrete : annuity due (True) or continuous (False)
        """
        u = self.max_term(x+s, u)
        if u < 0:
            return 0.
        a = self.deferred_annuity(x, s=s, u=u, t=t, b=b, discrete=discrete)
        return self.interest.annuity(u, m=int(discrete)) + a

    def increasing_annuity(self, x: int, s: int = 0, t: int = Insurance.WHOLE, 
                           b: int = 1, discrete: bool = True) -> float:
        """Increasing annuity

        Args:
          x : age of selection
          s : years after selection
          t : term of life annuity in years
          b : benefit amount at end of first year
          discrete : annuity due (True) or continuous (False)
        """
        t = self.max_term(x+s, t=t)
        benefit = lambda x, s: b * (s + 1) # increasing benefit
        return self.a_x(x, s=s, benefit=benefit, t=t, discrete=discrete)

    def decreasing_annuity(self, x: int, s: int = 0, t: int = 0, 
                           b: int = 1, discrete: bool = True) -> float:
        """Identity (Da)_x:n + (Ia)_x:n = (n+1) a_x:n temporary annuity

        Args:
          x : age of selection
          s : years after selection
          t : term of life annuity in years
          b : benefit amount at end of first year
          discrete : annuity due (True) or continuous (False)
        """
        assert t >= 0   # decreasing must be till term
        t = self.max_term(x+s, t=t)
        a = self.temporary_annuity(x, s=s, t=t, discrete=discrete)
        n = t + int(discrete)
        return b * (a*n - self.increasing_annuity(x, s=s, t=t, discrete=discrete))

    #
    # Annuity random variable: Y(t)
    #
    def Y_t(self, x: int, prob: float, discrete: bool = True) -> float:
        """T_x given percentile of the r.v. Y = PV of WL or Temporary Annuity

        Args:
          x : age of selection
          prob : desired probability threshold
          discrete : annuity due (True) or continuous (False)
        """
        assert prob < 1.0
        t = Insurance.solve(lambda t: self.S(x, 0, t), target=1-prob, grid=25)
        return math.floor(t) if discrete else t   # opposite of insurance

    def Y_from_t(self, t: float, discrete: bool = True) -> float:
        """PV of insurance payment Y(t), given T_x (or K_x if discrete)

        Args:
          t : year of death
          discrete : annuity due (True) or continuous (False)
        """
        if discrete:
            return (1 - self.interest.v_t(math.floor(t) + 1)) / self.interest.d
        else:
            return (1 - self.interest.v_t(t)) / self.interest.delta

    def Y_to_t(self, Y: float) -> float:
        """T_x  s.t. PV of annuity payments is Y

        Args:
          Y : Present value of benefits paid
        """
        t = math.log(1 - self.interest.delta * Y) / math.log(self.interest.v)
        return t

    def Y_from_prob(self, x: int, prob: float, discrete: bool = True) -> float:
        """Percentile of annuity PV r.v. Y, given probability

        Args:
          x : age initially insured
          prob : desired probability threshold
          discrete : annuity due (True) or continuous (False)
        """
        t = self.Y_t(x, prob, discrete=discrete)
        return self.Y_from_t(t=t, discrete=discrete)

    def Y_to_prob(self, x: int, Y: float) -> float:
        """Cumulative density of insurance PV r.v. Y, given percentile value

        Args:
          x : age initially insured
          Y : present value of benefits paid
        """
        t = self.Y_to_t(Y)
        return 1 - self.S(x, 0, t)   # opposite of Insurance

    def Y_x(self, x, s: int = 0, t: int = 1, discrete: bool = True) -> float:
        """EPV of year t annuity benefit for life aged [x]+s: b_x[s]+s(t)

        Args:
          x : age initially insured
          s : years after selection
          t : year of benefit
          discrete : annuity due (True) or continuous (False)
        """
        assert t >= 0
        if discrete:
            return (self.interest.v_t(math.floor(t)) 
                    * self.p_x(x, s=s, t=math.floor(t)))
        else:
            return (self.interest.v_t(t) * self.p_r(x, s=s, t=t))

    def Y_plot(self, x: int, s: int = 0, stop : int = 0,
               benefit: Callable = lambda x,k: 1,
               T: float | None = None,
               discrete: bool = True,
               ax: Any = None,
               title: str | None = None,               
               color='r') -> float:
        """Plot PV of annuity r.v. Y vs T

        Args:
          x : age of selection
          s : years after selection
          stop : time to end plot
          benefit : benefit as a function of selection age and time
          discrete : discrete or continuous insurance
          ax : figure object to plot in
          color : color to plot
          title : title of plot
        """
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        K = 'K' if discrete else 'T'
        stop = stop or self._MAXAGE - (x + s)
        step = 1 if discrete else stop / 1000.
        steps = np.arange(0, stop + step, step)

        # plot benefit values
        y = [benefit(x, s+t) * self.Y_from_t(t, discrete=discrete) for t in steps]
        if T is None:
            ax.bar(steps, y, width=stepsize, alpha=0.5, color=color)
            Y = None
        else:
            ax.step(steps, y, ':', c=color, where='pre' if discrete else 'post')

            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            yjig = (ymax - ymin) / 50
            xjig = (xmax - xmin) / 50
            
            # indicate PV benefit Y(T*)
            Y = self.Y_from_t(T, discrete=discrete) * benefit(self._MINAGE, T)
            ax.plot(T, Y, c=color, marker='o')
            ax.text(T + xjig, Y + 1*yjig, f"Y*={Y:.2f}", c=color)

            # indicate given T* time of death
            ax.vlines(T, ymin, Y, colors='g', linestyles=':')
            ax.text(T + xjig, ymin, f"{K}={T:.2f}", c='g')

            # indicate corresponding probability S(T*)
            p = 1 - self.S(x, 0, T)     # note this is opposite of insurance
            ax.hlines(Y, T, xmax, colors='g', linestyles=':')
            ax.text(xmax, Y - yjig, f"Prob={p:.3f}", c='g', va='top', ha='right')
        ax.set_title(f"PV annuity r.v. $Y({K}_{{{x if x else 'x'}}})$"
                     if title is None else title)
        ax.set_ylabel(f"$Y({K}_x)$", color=color)
        ax.set_xlabel(f"${K}_x$")
        plt.tight_layout()
        return Y

if __name__ == "__main__":
    from actuarialmath.sult import SULT
    life = SULT()
    life.Y_plot(x=65, T=life.Y_t(x=65, prob=0.5))

    
    life = Annuity().set_interest(delta=0.06)\
                    .set_survival(mu=lambda *x: 0.04)
    prob = 0.8
    x = 0
    discrete = False
    t = life.Y_t(0, prob, discrete=discrete)
    Y = life.Y_from_prob(x, prob=prob, discrete=discrete)
    print(t, life.Y_to_t(Y))
    print(Y, life.Y_from_t(t, discrete=discrete))
    print(prob, life.Y_to_prob(x, Y=Y))
    life.Y_plot(x=0, T=t, discrete=discrete)

    print("SOA Question 5.6:  (D) 1200")
    life = Annuity().set_interest(i=0.05)
    var = life.annuity_variance(A2=0.22, A1=0.45)
    mean = life.annuity_twin(A=0.45)
    print(life.portfolio_percentile(mean=mean, variance=var, prob=.95, N=100))
    print()
    
    print("Other usage")
    mu = 0.04
    delta = 0.06
    life = Annuity().set_interest(delta=delta)\
                    .set_survival(mu=lambda *x: mu)
    print(life.temporary_annuity(50, t=20, b=10000, discrete=False))
    print(life.endowment_insurance(50, t=20, b=10000, discrete=False))
    print(life.E_x(50, t=20))
    print(life.whole_life_annuity(50, b=10000, discrete=False))
    print(life.whole_life_annuity(70, b=10000, discrete=False))

    mu = 0.07
    delta = 0.02
    life = Annuity().set_interest(delta=delta)\
                    .set_survival(mu=lambda *x: mu)
    print(life.whole_life_annuity(0, discrete=False) * 30)   # 333.33
    print(life.temporary_annuity(0, t=10, discrete=False) * 30)  # 197.81
    print(life.interest.annuity(5, m=0))  # 4.7581
    print(life.deferred_annuity(0, u=5, discrete=False)) # 7.0848
    print(life.certain_life_annuity(0, u=5, discrete=False))  # 11.842

    mu = 0.02
    delta = 0.05
    life = Annuity().set_interest(delta=delta)\
                    .set_survival(mu=lambda *x: mu)
    print(life.decreasing_annuity(0, t=5, discrete=False))  # 6.94



"""Policy Values - Computes present value of future losses and reserves

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
import numpy as np
import scipy
import matplotlib.pyplot as plt
from typing import Dict, Any
from actuarialmath.premiums import Premiums
from actuarialmath.life import Actuarial

class Contract(Actuarial):
    """Set and retrieve policy contract terms

    Args:
      premium : level premium amount
      benefit : insurance death benefit amount
      settlement_policy : settlement expense per policy
      endowment : endowment benefit amount
      initial_policy : first year total expense per policy
      initial_premium : first year total premium per $ of gross premium
      renewal_policy : renewal expense per policy
      renewal_premium : renewal premium per $ of gross premium
      discrete : annuity due (True) or continuous (False)
      T : term of insurance
      discrete : annuity due (True) or continuous (False)        
    """
    
    def __init__(self, 
                 premium: float = 1,
                 initial_policy: float = 0, 
                 initial_premium: float = 0, 
                 renewal_policy: float = 0, 
                 renewal_premium: float = 0,
                 settlement_policy: float = 0,
                 benefit: float = 1, 
                 endowment: float = 0,
                 T: int = Premiums.WHOLE,
                 discrete: bool = True):
        self.premium=premium
        self.benefit=benefit
        self.discrete=discrete
        self.initial_policy=initial_policy
        self.initial_premium=initial_premium
        self.renewal_policy=renewal_policy
        self.renewal_premium=renewal_premium
        self.settlement_policy=settlement_policy
        self.endowment=endowment
        self.T = T
    
    def set_contract(self, **terms) -> Any:
        """Update any existing policy contract terms

        Args:
          **kwargs : one or more contract policy terms, and its value to update
        """
        for key, value in terms.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    @property
    def premium_terms(self) -> Dict:    # terms required by gross_premiums 
        """Getter returns dict of terms required for calculating gross premiums"""
        return dict(benefit=self.benefit,
                    initial_policy=self.initial_policy,
                    initial_premium=self.initial_premium,
                    renewal_policy=self.renewal_policy,
                    renewal_premium=self.renewal_premium,
                    settlement_policy=self.settlement_policy)

    def renewals(self, t: int = 0) -> "Contract":
        """Returns contract object with initial terms set to renewal terms

        Args:
          t : number of years after initial
        """
        return Contract(benefit=self.benefit,
                        renewal_policy=self.renewal_policy,
                        renewal_premium=self.renewal_premium,
                        initial_policy=self.renewal_policy,
                        initial_premium=self.renewal_premium,
                        settlement_policy=self.settlement_policy,
                        discrete=self.discrete,
                        premium=self.premium,
                        endowment=self.endowment,
                        T=self.T - t)

    @property
    def renewal_profit(self) -> float:
        """Getter returns renewal dollar profit (premium less renewal expenses)"""
        # premium less renewal per premium and expense"""
        return ((self.premium * (1 - self.renewal_premium)) - self.renewal_policy)

    @property
    def initial_cost(self) -> float:
        """Getter returns total initial cost (excludes renewal expenses)"""
        return ((self.initial_policy - self.renewal_policy) +
                (self.premium * (self.initial_premium - self.renewal_premium)))

    @property
    def claims_cost(self) -> float:
        """Getter returns total claims cost (death benefit + settlement expense)"""
        return self.benefit + self.settlement_policy

class PolicyValues(Premiums):
    """Compute net and gross future losses and policy values"""
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    #
    # Net Future Loss shortcuts for WL and Endowment Insurance 
    #
    def net_future_loss(self, A: float, A1: float, b: int = 1) -> float:
        """Shortcuts for WL or Endowment Insurance net loss

        Args:
          A : insurance factor at age (x)
          A1 : insurance factor at t years after x
          b : benefit amount
        """
        return b * (A1 - A) / (1 - A)

    def net_variance_loss(self, A1: float, A2: float, A: float = 0, 
                         b: int = 1) -> float:
        """Shortcuts for variance of net loss of WL or Endowment Insurance

        Args:
          A : insurance factor at age (x)
          A1 : first moment of insurance factor at t years after x
          A2 : insurance factor at double force of interest t years after x
          b : benefit amount
        """
        if not A:
            A = A1   # assume t = 0 => A = A1
        return b**2 * (A2 - A1**2) / (1 - A)**2

    def net_policy_variance(self, x, s: int = 0, t: int = 0, b: int = 1, 
                            n: int = Premiums.WHOLE, endowment: int = 0, 
                            discrete: bool = True) -> float:
        """Variance of future loss for WL or Endowment Ins assuming equivalence

        Args:
          x : age of selection
          s : years after selection
          t : term of life annuity in years
          n : number of years premiums paid
          b : benefit amount
          endowment : endowment amount
          discrete : annuity due (True) or continuous (False)
        """
        if n < 0:             # Whole Life
            A2 = self.whole_life_insurance(x, s=s+t, moment=2,
                                            discrete=discrete)
            A1 = self.whole_life_insurance(x, s=s+t, discrete=discrete)
            A = self.whole_life_insurance(x, s=s, discrete=discrete)
        elif endowment == b:   # Endowment Insurance 
            n = self.max_term(x+s+t, n)
            A2 = self.endowment_insurance(x, s=s+t, t=n-t, moment=2,
                                            discrete=discrete)
            A1 = self.endowment_insurance(x, s=s+t, t=n-t, discrete=discrete)
            A = self.endowment_insurance(x, s=s, t=n, discrete=discrete)
        else:
            raise Exception("Variances for WL and Endowment Ins only")
        return self.net_variance_loss(A=A, A1=A1, A2=A2, b=b)

    #
    # Net Policy Value for special and WL/endowment insurance
    #
    def net_policy_value(self, x: int, s: int = 0, t: int = 0, b: int = 1, 
                         n: int = Premiums.WHOLE, endowment: int = 0, 
                         discrete: bool = True) -> float:
        """Net policy value assuming premiums from equivalence: E[L_t]

        Args:
          x : age initially insured
          s : years after selection
          n : number of years of premiums paid
          t : year of death
          b : benefit amount
          endowment : endowment amount
          discrete : discrete/annuity due (True) or continuous (False)
        """
        if n < 0:             # Shortcut available for Whole Life
            A1 = self.whole_life_insurance(x, s=s+t, discrete=discrete)
            A = self.whole_life_insurance(x, s=s, discrete=discrete)
        elif endowment == b:  # Shortcut available for (equal) Endowment Insurance
            n = self.max_term(x+s+t, n)
            A1 = self.endowment_insurance(x, s=s+t, t=n-t, 
                                            discrete=discrete)
            A = self.endowment_insurance(x, s=s, t=n, discrete=discrete)
        else:   # Special Term or (unequal) Endowment insurance has no shortcut
            n = self.max_term(x+s+t, n)
            A1 = self.endowment_insurance(x, s=s+t, t=n-t, discrete=discrete,
                                            b=b, endowment=endowment)
            a1 = self.temporary_annuity(x, s=s+t, t=n-t, b=b, 
                                        discrete=discrete)
            A = self.endowment_insurance(x, s=s, t=n, discrete=discrete,
                                            b=b, endowment=endowment)
            a = self.temporary_annuity(x, s=s, t=n, b=b, discrete=discrete)
            return A1 - a1 * (A / a)
        return self.net_future_loss(A=A, A1=A1, b=b)  # apply shortcut

    #
    # Gross Future Loss shortcuts for WL and Endowment Insurance 
    #
    def gross_future_loss(self, A: float | None = None, 
                          a: float | None = None, 
                          contract: Contract | None = None) -> float:
        """Shortcut for WL or Endowment Insurance gross future loss

        Args:
          A : insurance factor at age (x)
          a : annuity factor at age (x)
          contract : policy contract terms and expenses
        """
        contract = contract or Contract()
        if a is None:    # assume WL or Endowment Insurance for twin annuity
            a = self.annuity_twin(A, discrete=contract.discrete)
        elif A is None:  # assume WL or Endowment Insurance for twin annuity
             A = self.insurance_twin(a, discrete=contract.discrete)
        return ((A * contract.claims_cost + contract.initial_cost) -
                (a * contract.renewal_profit))

    def gross_variance_loss(self, A1: float, A2: float = 0,
                            contract: Contract | None = None) -> float:
        """Shortcuts for variance of gross loss for WL or endowment insurance

        Args:
          A1 : insurance factor
          A2 : insurance factor at double the force of interest
          policy : policy terms and expenses
        """
        contract = contract or Contract()
        interest = self.interest.d if contract.discrete else self.interest.delta
        return (((contract.renewal_profit / interest) + contract.benefit +
                 contract.settlement_policy)**2 * (A2 - A1**2))

    def gross_policy_variance(self, x: int, s: int = 0, t: int = 0,
                              n: int = Premiums.WHOLE,
                              contract: Contract | None = None) -> float:
        """Variance of gross policy value for WL and Endowment Insurance

        Args:
          x : age initially insured
          s : years after selection
          n : number of years of premiums paid
          t : year of death
          contract : policy contract terms and expenses
        """
        contract = contract or Contract()
        if n < 0:  # WL
            A2 = self.whole_life_insurance(x, s=s+t, moment=2,
                                            discrete=contract.discrete)
            A1 = self.whole_life_insurance(x, s=s+t, 
                                            discrete=contract.discrete)
        elif contract.endowment == contract.claims_cost:  # Endowment
            n = self.max_term(x+s+t, n)
            A2 = self.endowment_insurance(x, s=s+t, t=n-t, moment=2, 
                                            discrete=contract.discrete)
            A1 = self.endowment_insurance(x, s=s+t, t=n-t, 
                                            discrete=contract.discrete)
        else:
            raise Exception("Variance for WL or Endowment Ins only")
        return self.gross_variance_loss(A1=A1, A2=A2, contract=contract)

    #
    # Gross Policy Value for special insurance
    #
    def gross_policy_value(self, x: int, s: int = 0, t: int = 0,
                           n: int = Premiums.WHOLE,
                           contract: Contract | None = None) -> float:
        """Gross policy values for insurance: t_V = E[L_t]

        Args:
          x : age initially insured
          s : years after selection
          t : number of years of premiums paid
          n : term of insurance
          contract : policy contract terms
        """
        contract = contract or Contract()
        if n < 0:    # Whole life shortcut
            A = self.whole_life_insurance(x, s=s+t, 
                                          discrete=contract.discrete)
        elif contract.endowment == contract.claims_cost:  # Endowment Ins shortcut
            n = self.max_term(x+s+t, n)
            A = self.endowment_insurance(x, s=s+t, t=n-t, 
                                            discrete=contract.discrete)
        else:  # Special term insurance
            n = self.max_term(x+s+t, n)
            A = self.term_insurance(x, s=s+t, t=n-t, discrete=contract.discrete)
            a = self.temporary_annuity(x, s=s+t, t=n-t,
                                       discrete=contract.discrete)
            endowment = 0
            if contract.endowment:  # endowment not equal to claims cost
                endowment = self.E_x(x, s=s+t, t=n-t) * contract.endowment
            initial_cost = 0 if t else contract.initial_cost
            return (A * contract.claims_cost + initial_cost + endowment -
                    (a * contract.renewal_profit))
        contract = contract.renewals(t) if t else contract # ignore initial if t>0
        return self.gross_future_loss(A=A, contract=contract)

    #
    # Future Loss random variable: L(T_x)
    #
    def L_from_t(self, t: float, contract: Contract | None = None) -> float:
        """PV of Loss L(t) at time of death t = T_x (or K_x if discrete)

        Args:
          t : year of death
          contract : policy contract
        """
        contract = contract or Contract()
        k = math.floor(t) if contract.discrete else t  # if endowment paid
        if contract.T > 0 and k >= contract.T:
            t = contract.T
            endowment = contract.endowment * self.Z_from_t(t)
        else:
            endowment = 0
        return ((contract.claims_cost
                 * self.Z_from_t(t, discrete=contract.discrete))
                + contract.initial_cost
                + endowment
                - (contract.renewal_profit
                   * self.Y_from_t(t, discrete=contract.discrete)))

    def L_to_t(self, L: float, contract: Contract | None = None) -> float:
        """Compute time of death T_x s.t. PV future loss is L

        Args:
          L : PV of future loss
          contract : policy contract terms and expenses
        """
        contract = contract or Contract()
        return Contract.solve(lambda t: self.L_from_t(t, contract),
                              target=L, grid=(0, self._MAXAGE), mad=True)

    def L_from_prob(self, x: int, prob: float, 
                    contract: Contract | None = None) -> float:
        """Compute PV of future loss at given percentile prob

        Args:
          x : age
          prob : probability threshold
          contract : policy contract
        """
        t = self.Z_t(x, prob, discrete=contract.discrete)
        return self.L_from_t(t, contract)

    def L_to_prob(self, x: int, L: float,
                  contract: Contract = Contract()) -> float:
        """Compute percentile of L on the PV of future loss curve"

        Args:
          x : age selected
          L : PV of future loss
          contract : policy contract terms and expenses
        """
        t = self.L_to_t(L, contract)
        return self.S(x, 0, t)

    def L_plot(self, x: int, s: int = 0, stop: int = 0,
               T: float | None = None,
               contract: Contract = Contract(),
               ax: Any = None,
               title: str | None = None,
               color='r') -> float:
        """Plot PV of future loss r.v. L vs time of death T_x
        
        Args:
          x : age selected
          s : years after selection
          stop : time to end plot
          contract : policy contract terms and expenses
          T : point in time to indicate probability and loss values
          title : title of plot 
          color : color to plot curve
        """        
        if ax is None:
            fig, ax = plt.subplots(1, 1)
        K = 'K' if contract.discrete else 'T'
        stop = stop or self._MAXAGE - (x + s)
        step = 1 if contract.discrete else stop / 1000.
        steps = np.arange(0, stop + step, step)

        # plot PV loss values
        y = [self.L_from_t(t, contract=contract) for t in steps]
        ax.plot(steps, y, '.', c=color)
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        yjig = (ymax - ymin) / 50
        xjig = (xmax - xmin) / 50
        if T is None:
            ax.plot(steps, y, ".", color=color)
            ax.hlines(0, xmin, xmax, colors='g', linestyles=':')
            z = None
        else:
            # plot indicate(T*)
            z = self.L_from_t(T, contract=contract)
            ax.plot(T, z, c=color, marker='o')
            ax.text(T, z + yjig, f"L*={z:.2f}", c=color)

            # indicate given time of death T*
            ax.vlines(T, ymin, z, colors='g', linestyles=':')
            ax.text(T + xjig, ymin, f"{K}={T:.2f}", c='g')

            # indicate corresponding S(T*)
            p = self.S(x, s, T)
            ax.hlines(z, T, xmax, colors='g', linestyles=':')
            ax.text(xmax, z-yjig, f"Prob={p:.3f}", c='g', va='top', ha='right')
            #ax.set_title(f"Pr[${K}_x$ >= {K}(Z*)] > {p:.3}")
        ax.set_title(f"PV future loss r.v. $L({K}_{{{x if x else 'x'}}})$"
                     if title is None else title)
        ax.set_ylabel(f"$L({K}_x)$", color=color)
        ax.set_xlabel(f"${K}_x$")
        plt.tight_layout()
        return z

if __name__ == "__main__":
    from actuarialmath.sult import SULT
    from actuarialmath.policyvalues import Contract
    life = SULT()
    x = 20
    P = life.net_premium(x=x)
    contract = Contract(premium=P, discrete=True)
    T = life.L_to_t(L=0, contract=contract)  # breakeven T
    print(T)
    life.L_plot(x=x, T=T, contract=contract)

    print("SOA Question 6.24:  (E) 0.30")
    life = PolicyValues().set_interest(delta=0.07)
    x, A1 = 0, 0.30   # Policy for first insurance
    P = life.premium_equivalence(A=A1, discrete=False)  # Need its premium
    contract = Contract(premium=P, discrete=False)
    def fun(A2):  # Solve for A2, given Var(Loss)
        return life.gross_variance_loss(A1=A1, A2=A2, contract=contract)
    A2 = life.solve(fun, target=0.18, grid=0.18)
    contract = Contract(premium=0.06, discrete=False)     # Solve second insurance
    variance = life.gross_variance_loss(A1=A1, A2=A2, contract=contract)
    print(variance)
    print()

    print("SOA Question 6.30:  (A) 900")
    life = PolicyValues().set_interest(i=0.04)
    contract = Contract(premium=2.338, benefit=100, initial_premium=.1,
                        renewal_premium=0.05)
    var = life.gross_variance_loss(A1=life.insurance_twin(16.50),
                                   A2=0.17, contract=contract)
    print(var)
    print()

    print("SOA Question 7.32:  (B) 1.4")
    life = PolicyValues().set_interest(i=0.06)
    contract = Contract(benefit=1, premium=0.1)
    def fun(A2):
        return life.gross_variance_loss(A1=0, A2=A2, contract=contract)
    A2 = life.solve(fun, target=0.455, grid=0.455)
    contract = Contract(benefit=2, premium=0.16)
    var = life.gross_variance_loss(A1=0, A2=A2, contract=contract)
    print(var)
    print()

    print("SOA Question 6.12:  (E) 88900")
    life = PolicyValues().set_interest(i=0.06)
    a = 12
    A = life.insurance_twin(a)
    contract = Contract(benefit=1000, settlement_policy=20, 
                        initial_policy=10, initial_premium=0.75, 
                        renewal_policy=2, renewal_premium=0.1)
    contract.premium = life.gross_premium(A=A, a=a, **contract.premium_terms)
    print(A, contract.premium)
    L = life.gross_variance_loss(A1=A, A2=0.14, contract=contract)
    print(L)
    print()

    from actuarialmath.sult import SULT
    print("SOA Question 6.6:  (B) 0.79")
    life = SULT()
    P = life.net_premium(62, b=10000)
    contract = Contract(premium=1.03*P, renewal_policy=5,
                        initial_policy=5, initial_premium=0.05, benefit=10000)
    L = life.gross_policy_value(62, contract=contract)
    var = life.gross_policy_variance(62, contract=contract)
    prob = life.portfolio_cdf(mean=L, variance=var, value=40000, N=600)
    print(prob, 0.79)
    life.L_plot(62, contract=contract)
    print()

    print("SOA Question 7.6:  (E) -25.4")
    from actuarialmath.sult import SULT
    life = SULT()
    P = life.net_premium(45, b=2000)
    contract = Contract(benefit=2000, initial_premium=.25, renewal_premium=.05,
                        initial_policy=2*1.5 + 30, renewal_policy=2*.5 + 10)
    G = life.gross_premium(a=life.whole_life_annuity(45),
                           **contract.premium_terms)
    gross = life.gross_policy_value(45, t=10,
                                    contract=contract.set_contract(premium=G))
    net = life.net_policy_value(45, t=10, b=2000)
    V = gross - net
    print(V, -25.4)
    T = life.L_to_t(G, contract=contract)
    print(G)

    



"""Constant Force of Mortality - shortcut formulas

MIT License. Copyright 2022-2023 Terence Lim
"""
import math
from scipy.stats import norm
from actuarialmath.mortalitylaws import MortalityLaws

class ConstantForce(MortalityLaws):
    """Constant force of mortality - memoryless exponential distribution of lifetime

    Args:
      mu : constant value of force of mortality
      udd : assume UDD (True) or CFM (False, default) between integer ages
    """

    def __init__(self, mu: float, udd: bool = False, **kwargs):
        super().__init__(udd=udd, **kwargs)

        def _mu(x: int, s: float) -> float:
            """Constant force of mortality"""
            return mu

        def _S(x: int, s, t: float) -> float: 
            """Shortcut for survival function with constant force of mortality"""
            return math.exp(-mu * t)

        self.set_survival(mu=_mu, S=_S)
        self.mu_ = mu   # store mu parameter

    def e_x(self, x: int, s: int = 0, t: int = MortalityLaws.WHOLE, 
           curtate: bool = False, moment: int = 1) -> float:
        """Expected lifetime E[T_x] is memoryless: does not depend on (x)

        Args:
          x : age of selection
          s : years after selection
          t : limited at year t
          curtate : whether curtate (True) or continuous (False) lifetime
          moment : first (1) or second (2) moment
        """
        if not curtate:    # Var[Tx] = 1/mu^2, E[Tx] = 1/mu
            if moment == MortalityLaws.VARIANCE:
                return 1. / self.mu_**2   # shortcut
            elif moment == 1:
                e = 1 / self.mu_  # infinite n case shortcut
                if t >= 0:        # n-limited from recursion e_x=e_x|n + n_p_x e_x
                    e *= (1 - math.exp(-self.mu_ * t))
                return e
        return super().e_x(x, s=s, t=t, curtate=curtate, moment=moment)

    def E_x(self, x: int, s: int = 0, t: int = MortalityLaws.WHOLE, 
            endowment: int = 1, moment: int = 1) -> float:
        """Shortcut for pure endowment: does not depend on age x

        Args:
          x : age of selection
          s : years after selection
          t : term of pure endowment
          endowment : amount of pure endowment
          moment : compute first or second moment
        """
        if t == 0:
            return 1.
        if t < 0:
            return 0.
        delta = moment * self.interest.delta   # multiply force of interest
        return math.exp(-(self.mu_ + delta) * t) * endowment**moment

    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1,
                           variance: bool = False,
                           discrete: bool = True) -> float:
        """Shortcut for whole life annuity: does not depend on age x

        Args:
          x : age of selection
          s : years after selection
          b : annuity benefit amount
          variance : return APV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.whole_life_insurance(x, s=s, discrete=discrete)
            A2 = self.whole_life_insurance(x, s=s, moment=2, discrete=discrete)
            return (b**2 * (A2 - A1**2) /
                    (self.interest.d if discrete else self.interest.delta)**2)
        if not discrete:
            return (1. / (self.mu_ + self.interest.delta)) * b
        return super().whole_life_annuity(x, s=s, b=b, discrete=discrete)

    def temporary_annuity(self, x: int, s: int = 0, t: int = MortalityLaws.WHOLE,
                          b: int = 1, variance: bool = False,
                          discrete: bool = True) -> float:
        """Shortcut for temporary life annuity: does not depend on age x

        Args:
          x : age of selection
          s : years after selection
          t : term of annuity in years
          b : annuity benefit amount
          variance : return APV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        interest = self.interest.d if discrete else self.interest.delta
        if variance:  # short cut for variance of temporary life annuity
            A1 = self.endowment_insurance(x, s=s, t=t, discrete=discrete)
            A2 = self.endowment_insurance(x, s=s, t=t, moment=2,
                                          discrete=discrete)
            return (b**2 * (A2 - A1**2) /
                    (self.interest.d if discrete else self.interest.delta)**2)
        if not discrete:
            a = b * 1/(self.mu_ + self.interest.delta)
            if t < 0:
                return a
            else:
                return a * (1 - math.exp(-(self.mu_ + self.interest.delta)*t))
        return super().temporary_annuity(x, s=s, b=b, t=t, discrete=discrete)

    def whole_life_insurance(self, x: int, s: int = 0, moment: int = 1,
                             b: int = 1, discrete: bool = True) -> float:
        """Shortcut for APV of whole life: does not depend on age x

        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if moment > 0 and not discrete:
            delta = moment * self.interest.delta   # multiply force of interest
            return (self.mu_ / (self.mu_ + delta))
        return super().whole_life_insurance(x, s=s, moment=moment, b=b,
                                            discrete=discrete)

    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1,
                       moment: int = 1, discrete: bool = True) -> float:
        """Shortcut for APV of term life: does not depend on age x

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        if moment > 0 and not discrete:
            delta = moment * self.interest.delta   # multiply force of interest
            A = b**moment * self.mu_/(self.mu_ + delta)
            if t < 0:
                return A
            else:
                return A * (1 - math.exp(-(self.mu_ + delta)*t))
        return super().term_insurance(x, s=s, t=t, b=b, moment=moment,
               discrete = discrete)

    def Z_t(self, x: int, prob: float, discrete: bool = True) -> float:
        """Shortcut for T_x (or K_x) given survival probability for insurance

        Args:
          x : age selected
          prob : desired probability threshold
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        t = -math.log(prob) / self.mu_
        return math.floor(t) if discrete else t    # opposite of annuity        

    def Y_t(self, x: int, prob: float, discrete: bool = True) -> float:
        """Shortcut for T_x (or K_x) given survival probability for annuity

        Args:
          x : age selected
          prob5~ : desired probability threshold
          discrete : continuous (False) or annuity due (True)
        """
        t = -math.log(1 - prob) / self.mu_
        return math.ceil(t) if discrete else t    # opposite of insurance

if __name__ == "__main__":
    print("SOA Question 6.36:  (B) 500")
    life = ConstantForce(mu=0.04).set_interest(delta=0.08)
    a = life.temporary_annuity(50, t=20, discrete=False)
    A = life.term_insurance(50, t=20, discrete=False)
    print(a,A)
    def fun(R):
       return life.gross_premium(a=a, A=A, initial_premium=R/4500,
                                 renewal_premium=R/4500, benefit=100000)
    R = life.solve(fun, target=4500, grid=[400, 800])
    print(R)
    print()

    print("SOA Question 6.31:  (D) 1330")
    life = ConstantForce(mu=0.01).set_interest(delta=0.05)
    A = life.term_insurance(35, t=35) + life.E_x(35, t=35) * 0.51791 # A_35
    A = (life.term_insurance(35, t=35, discrete=False) 
            + life.E_x(35, t=35) * 0.51791)    # A_35
    P = life.premium_equivalence(A=A, b=100000, discrete=False)
    print(P)
    print()

    print("SOA Question 6.27:  (D) 10310")
    life = ConstantForce(mu=0.03).set_interest(delta=0.06)
    x = 0
    payments = (3 * life.temporary_annuity(x, t=20, discrete=False) 
            + life.deferred_annuity(x, u=20, discrete=False))
    benefits = (1000000 * life.term_insurance(x, t=20, discrete=False)
            + 500000 * life.deferred_insurance(x, u=20, discrete=False))
    print(benefits, payments)
    print(life.term_insurance(x, t=20), life.deferred_insurance(x, u=20))
    P = benefits / payments
    print(P)
    print()


    print("SOA Question 5.4:  (A) 213.7")
    life = ConstantForce(mu=0.02).set_interest(delta=0.01)
    P = 10000 / life.certain_life_annuity(40, u=life.e_x(40, curtate=False), 
                                          discrete=False)
    print(P)
    print()


    print("SOA Question 5.1: (A) 0.705")
    life = ConstantForce(mu=0.01).set_interest(delta=0.06)
    EY = life.certain_life_annuity(0, u=10, discrete=False)
    print(life.p_x(0, t=life.Y_to_t(EY)))  # 0.705
    print()




"""Recursion - Applies recursion, shortcut and actuarial formulas

MIT License. Copyright 2022-2023 Terence Lim
"""
from typing import Tuple
from actuarialmath.reserves import Reserves

class Recursion(Reserves):
    """Solve by appling recursive, shortcut and actuarial formulas repeatedly

    Args:
      depth : maximum depth of recursions (default is 3)
      verbose : whether to echo recursion steps (True, default)

    Notes:
      7 types of information can be loaded and calculated in recursions:

      - 'q' : (deferred) probability (x) dies in t years
      - 'p' : probability (x) survives t years
      - 'e' : (temporary) expected future lifetime, and moments
      - 'A' : deferred, term, endowment or whole life insurance, and moments
      - 'IA' : decreasing life insurance of t years
      - 'DA' : increasing life insurance of t years
      - 'a' : deferred, temporary or whole life annuity of t years, and moments
    """

    def __init__(self, depth: int = 3, verbose: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.db = {}
        self.maxdepth = depth
        Blog._verbose = verbose

    #
    # helpers to store given input values
    #
    def db_key(self, *args, **kwargs) -> Tuple:
        """Generate a unique key representing values of given arguments"""
        assert args and kwargs
        return tuple(list(args) + sorted(kwargs.items()))

    def db_put(self, key: Tuple, value: float | None) -> "Recursion":
        """Store the item's key and value; or remove if value is None
        Args:
          key : key of the item
          value : value to store for item
        """
        if value is None and key in self.db:
            self.db.pop(key)
        else:
            self.db[key] = value
        return self

    def db_print(self):
        """Display the stored keys and values"""
        for k in sorted(self.db.keys()):
            print(k, self.db[k])

    #
    # Formulas for Mortality: u|t_q_x
    #
    def get_q(self, x: int, s: int = 0, t: int = 1,
              u: int = 0) -> float | None:
        """Get mortality rate from key-value store

        Args:
          x : age of selection
          s : years after selection
          u : survive u years, then...
          t : death within next t years        
        """
        key = self.db_key('q', x=x+s, u=u, t=t)
        return self.db.get(key, None)

    def set_q(self, val: float, x: int, s: int = 0, t: int = 1, 
              u: int = 0) -> "Recursion":
        """Set mortality rate u|t_q_[x]+s to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          u : survive u years, then...
          t : death within next t years        
        """
        return self.db_put(self.db_key('q', x=x+s, u=u, t=t), val)

    def _q_x(self, x: int, s: int = 0, t: int = 1, u: int = 0, 
             depth: int = 1) -> float:
        """Helper to compute mortality from recursive and alternate formulas"""
        found = self.get_q(x, s=s, t=t, u=u)
        if found is not None:
            return found
        if t == 0:
            return 0
        if t < 0:
            return 1
        if u > 0:
            pu = self._p_x(x, s=s, t=u, depth=depth-1) #1-abs(depth))
            qt = self.get_q(x, s=s+u, t=t)
            if pu is not None and qt is not None:
                self.blog(Blog.q(x=x, s=s, t=t, u=u), '=',
                          Blog.p(x=x, s=s, t=u), '*', Blog.q(x=x, s=s+u, t=t),
                          depth=depth, rule="defer mortality")
                return pu * qt        # (1) u_p_x * t_q_x+u
            qu = self.get_q(x, s=s, t=u)
            qt = self.get_q(x, s=s, t=u+t)
            if qu is not None and qt is not None:
                self.blog(Blog.q(x=x, s=s, t=t, u=u), '=',
                          Blog.q(x=x, s=s, t=t+u), '-', Blog.q(x=x, s=s, t=u),
                          depth=depth, rule="limit mortality")
                return qt - qu        # (2) u+t_q_x - u_q_x
        if depth <= 0:
            return None
        pu = self._p_x(x, s=s, t=u, depth=depth-1)
        pt = self._p_x(x, s=s, t=u+t, depth=depth-1)
        if pu is not None and pt is not None:
            self.blog(Blog.q(x=x, s=s, t=t, u=u), '=',
                      Blog.p(x=x, s=s, t=u), '-', Blog.p(x=x, s=s, t=t+u),
                      depth=depth, rule="complement survival")
            return pu - pt            # (3) u_p_x - u+t_p_x


    def q_x(self, x: int, s: int = 0, t: int = 1, u: int = 0) -> float:
        self.blog = Blog("Mortality",
                         Blog.q(x=x, s=s, t=t, u=u),
                         levels=self.maxdepth)
        """Compute mortality rate by calling recursion helper"""
        q = self._q_x(x, s=s, t=t, u=u, depth=self.maxdepth)
        if q is not None:
            print(self.blog, end='')
        return q

    #
    # Formulas for Survival: t_p_x
    #
    def get_p(self, x: int, s: int = 0, t: int = 1) -> float | None:
        """Get survival probability from key-value store

        Args:
          x : age of selection
          s : years after selection
          t : survives next t years
        """
        if t == 0:
            return 1
        if t < 0:
            return 0
        key = self.db_key('p', x=x+s, t=t)
        return self.db.get(key, None)

    def set_p(self, val: float, x: int, s: int = 0, t: int = 1) -> "Recursion":
        """Set survival probability t_p_[x]+s to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          t : survives next t years
        """
        return self.db_put(self.db_key('p', x=x+s, t=t), val)

    def _p_x(self, x: int, s: int = 0, t: int = 1, depth: int = 1) -> float:
        """Helper to compute survival from recursive and alternate formulas"""
        found = self.get_p(x, s=s, t=t)
        if found is not None:
            return found
        found = self.get_q(x, s=s, t=t)  
        if found is not None:
            self.blog(Blog.p(x=x, s=s, t=t), '= 1 -', Blog.q(x=x, s=s, t=t),
                      depth=depth, rule='complement of mortality')
            return 1 - found  # (1) complement of q_x
        if depth <= 0:
            return None
        
        # (2a) inverse chain rule: p_x(t) = p_x-1(t+1) / p_x-1 
        found = self._p_x(x, s=s-1, t=t+1, depth=depth-1)
        p = self._p_x(x, s=s-1, t=1, depth=1-abs(depth))
        if found is not None and p is not None:
            self.blog(Blog.p(x=x, s=s, t=t), '=',
                      Blog.p(x=x, s=s-1, t=t+1), '/', Blog.p(x=x, s=s-1, t=1),
                      depth=depth, rule="survival chain rule")
            return found / p
        
        # (2b) inverse chain rule: p_x(t) = p_x(t+1) / p_x+t 
        found = self._p_x(x, s=s, t=t+1, depth=depth-1)
        p = self._p_x(x, s=s+t, t=1, depth=1-abs(depth))
        if found is not None and p is not None:
            self.blog(Blog.p(x=x, s=s, t=t), '=',
                      Blog.p(x=x, s=s, t=t+1), '/', Blog.p(x=x, s=s+t, t=1),
                      depth=depth, rule="survival chain rule")
            return found / p
        if t > 1:
            # (3a) chain rule: p_x(t) = p_x * p_x+1(t-1)
            found = self._p_x(x, s=s+1, t=t-1, depth=depth-1)
            p = self._p_x(x, s=s, t=1, depth=1-abs(depth))
            if found is not None and p is not None:
                self.blog(Blog.p(x=x, s=s, t=t), '=',
                          Blog.p(x=x, s=s+1, t=t-1), '*', Blog.p(x=x, s=s, t=1),
                          depth=depth, rule="survival chain rule")
                return found * p

            # (3b) chain rule: p_x(t) = p_x+t-1 * p_x(t-1)
            found = self._p_x(x, s=s, t=t-1, depth=depth-1)
            p = self._p_x(x, s=s+t-1, t=1, depth=1-abs(depth))
            if found is not None and p is not None:
                self.blog(Blog.p(x=x, s=s, t=t), '=',
                          Blog.p(x=x, s=s, t=t-1), '*', Blog.p(x=x, s=s+t-1, t=1),
                          depth=depth, rule="survival chain rule")
                return found * p

        if t == 1:
            E = self._E_x(x, s=s, t=1, depth=depth-1)
            if E is not None:
                self.blog(Blog.p(x=x, s=s, t=1), '=',
                          Blog.E(x=x, s=s, t=1), "/v",
                          depth=depth, rule="one-year endowment")
                return E / self.interest.v

            # (4a) annuity recursion: p_x = [a_x(t) - 1] / [v a_x+1(t-1)
            for _t in [self.WHOLE, 2]:  # consider only WL and 2-term
                a = self._a_x(x, s=s, t=_t, depth=depth-1)
                a1 = self._a_x(x, s=s+1, t=self.add_term(_t, -1), depth=depth-1)
                if a is not None and a1 is not None:
                    self.blog(Blog.p(x=x, s=s, t=1), '= [',
                              Blog.a(x=x, s=s, t=_t), '- 1 ] / [ v *',
                              Blog.a(x=x, s=s+1, t=_t-1), ']',
                              depth=depth, rule="annuity recursion")
                    return (a - 1) / (self.interest.v * a1)

            
            # (4b) insurance recursion: p_x = [v - A_x(t)] / [v (1 - A_x+1(t-1))]
            for _t in [self.WHOLE, 2]:  # consider only WL and 2-term
                A = self._A_x(x, s=s, t=_t, depth=depth-1)
                A1 = self._A_x(x, s=s+1, t=self.add_term(_t, -1), depth=depth-1)
                if A is not None and A1 is not None:
                    self.blog(Blog.p(x=x, s=s, t=1), '= [ v -',
                              Blog.A(x=x, s=s, t=_t), '] / [v * [ 1 -',
                              Blog.A(x=x, s=s+1, t=_t-1), ']]',
                              depth=depth, rule="insurance recursion")
                    return (self.interest.v - A)/(self.interest.v * (1 - A1))

    def p_x(self, x: int, s: int = 0, t: int = 1) -> float:
        """Compute survival probability by calling recursion helper

        Args:
          x : age of selection
          s : years after selection
          t : survives at least t years
        """
        self.blog = Blog("Survival",
                         Blog.p(x=x, s=s, t=t),
                         levels=self.maxdepth)
        p = self._p_x(x, s=s, t=t, depth=self.maxdepth)
        if p is not None:
            print(self.blog, end='')
        return p

    #
    # Formulas for Expected Future Lifetime: e_x
    #
    def get_e(self, x: int, s: int = 0, t: int = Reserves._WHOLE, 
              curtate: bool = False, moment: int = 1) -> float | None:
        """Get expected future lifetime from key-value store

        Args:
          x : age of selection
          s : years after selection
          t : limit of expected future lifetime
          curtate : curtate (True) or complete expectation (False)
          moment : first or second moment of expected future lifetime
        """
        key = self.db_key('e', x=x+s, t=t, curtate=curtate, moment=moment)
        return self.db.get(key, None)

    def set_e(self, val: float, x: int, s: int = 0, t: int = Reserves._WHOLE, 
              curtate: bool = False, moment: int = 1) -> "Recursion":
        """Set expected future lifetime e_[x]+s:t to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          t : limit of expected future lifetime
          curtate : curtate (True) or complete expectation (False)
          moment : first or second moment of expected future lifetime
        """
        return self.db_put(self.db_key('e', x=x+s, t=t, moment=moment,
                                       curtate=curtate), val)

    def _e_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, 
            curtate: bool = False, moment: int = 1, 
            depth: int = 1) -> float | None:
        """Helper to compute from recursive and alternate formulas"""
        found = self.get_e(x, s=s, t=t, curtate=curtate, moment=moment)
        if found is not None:
            return found
        if depth <= 0:
            return None
        if moment == 1:
            if t > 0:  
                p_t = self._p_x(x, s=s, t=t)
                if t == 1 and curtate:
                    self.blog(Blog.e(x=x, s=s, t=1, curtate=curtate), '=',
                              Blog.p(x=x, s=s, t=1),
                              #f"e_{x+s}:1",
                              depth=depth, rule='1-year curtate shortcut')
                    return p_t   # (1) if curtate and t=1: e_x:1 = p_x 
                if t > 1:
                    e = self._e_x(x, s=s, t=Reserves._WHOLE, curtate=curtate, 
                                  moment=1, depth=depth-1)
                    e_t = self._e_x(x, s=s+t, t=Reserves._WHOLE, curtate=curtate,
                                    moment=1, depth=depth-1)
                    if e is not None and e_t is not None and p_t is not None:
                        self.blog(Blog.e(x=x, s=s, t=t), '=', Blog.e(x=x, s=s), '-',
                                  Blog.p(x=x, s=s), '*', Blog.e(x=x, s=s+t),
                                  #"e_{x+s}:{t}: e_x - p_x e_x+t",
                                  depth=depth, rule="temporary lifetime")
                        return e - p_t*e_t  # (2) temporary = e_x - t_p_x e_x+t
            e = self._e_x(x, s=s-1, t=self.add_term(t, 1), curtate=curtate,
                          moment=1, depth=depth-1)
            e1 = self._e_x(x, s=s-1, t=1, curtate=curtate, moment=1,
                          depth=depth-1)
            p = self._p_x(x, s=s-1)
            if e is not None and e1 is not None and p is not None:
                #_t = "" if t < 0 else ":" + str(t)
                #msg = f"forward e_{x+s}{_t} = e_{x+s}:1 + p_{x+s} e_{x+s+1}"
                self.blog(Blog.e(x=x, s=s, t=t), '=', Blog.e(x=x, s=s, t=t), '+',
                          Blog.p(x=x, s=s), '*', Blog.e(x=x, s=s+t),
                          depth=depth, rule='lifetime forward recursion')
                return (e - e1) / p # (3) forward: (e_x-1 - e_x-1:1) / p_x-1
            e = self._e_x(x, s=s, t=1, curtate=curtate,
                          moment=1, depth=depth-1)
            e_t = self._e_x(x, s=s+1, t=self.add_term(t, -1), curtate=curtate,
                            moment=1, depth=depth-1)
            p = self._p_x(x, s=s)
            if e is not None and e_t is not None and p is not None:
                self.blog(Blog.e(x=x, s=s, t=t), '=', Blog.e(x=x ,s=s, t=1), '+',
                          Blog.p(x=x, s=s), '*', Blog.e(x=x, s=s+1, t=t-1),
                          #f"backward: e_x:1 + p_x e_x+1:{t}",
                          depth=depth, rule='lifetime backward recursion')
                return e + p * e_t # (4) backward: e_x:1 + p_x e_x+1:t-1


    def e_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, 
            curtate: bool = False, moment: int = 1) -> float:
        """Compute expected future lifetime by calling recursion helper

        Args:
          x : age of selection
          s : years after selection
          t : limited at t years
          curtate : whether curtate (True) or complete (False) lifetime
          moment : whether to compute first (1) or second (2) moment
        """
        self.blog = Blog("Lifetime",
                         Blog.e(x=x, s=s, t=t, moment=moment, curtate=curtate),
                         levels=self.maxdepth)
        e = self._e_x(x, s=s, t=t, curtate=curtate, moment=moment,
                      depth=self.maxdepth)
        if e is not None:
            print(self.blog, end='')
            return e

    #
    # Formulas for Pure Endowment: t_E_x
    #
    def get_E(self, x: int, s: int = 0, t: int = 1, 
              endowment: int = 1, moment: int = 1) -> float | None:
        """Get pure endowment from key-value store

        Args:
          x : age of selection
          s : years after selection
          t : death within next t years
          endowment : endowment value
          moment : first or second moment of pure endowment
        """
        key = self.db_key('E', x=x+s, t=t, moment=moment)
        val = self.db.get(key, None)
        if val is not None:
            return val * endowment   # stored with benefit=1

    def set_E(self, val: float, x: int, s: int = 0, t: int = 1, 
              endowment: int = 1, moment: int = 1) -> "Recursion":
        """Set pure endowment t_E_[x]+s to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          t : death within next t years
          endowment : endowment value
          moment : first or second moment of pure endowment
        """
        val /= endowment   # store with benefit=1
        return self.db_put(self.db_key('E', x=x+s, t=t, moment=moment), val)

    def _E_x(self, x: int, s: int = 0, t: int = 1, endowment: int = 1, 
             moment: int = 1, depth: int = 1) -> float:
        """Helper to compute from recursive and alternate formulas"""
        E = self.get_E(x, s=s, t=t, endowment=endowment, moment=moment)
        if E is not None:
            return E
        if t < 0:     # t infinite => EPV(t) = 0
            return 0
        if t == 0:    # t = 0 => EPV(0) = endowment**moment
            return endowment**moment
        if moment > 1:
            E = self._E_x(x, s=s, endowment=endowment, depth=depth)
            if E:  # (1) Shortcut: 2E_x = v E_x
                self.blog(Blog.E(x=x, s=s, t=t, moment=moment),
                          f"= v(t={t})"+Blog.m(moment), '*', Blog.E(x=x, s=s, t=t),
                          depth=depth, rule='moments of pure endowment')
                return E * self.interest.v**(moment-1) 
        p = self._p_x(x, s=s, t=t, depth=depth-1)  # 1-abs(depth))
        if p is not None:   # (2) E_x = p_x * v
            #msg = f"pure endowment {t}_E_{x+s} = {t}_p_{x+s} * v^{t}"
            self.blog(Blog.E(x=x, s=s, t=t), '=', Blog.p(x=x, s=s, t=t),
                      f"* v(t={t})"+Blog.m(moment),
                      depth=depth, rule='pure endowment')
            return p * (endowment * self.interest.v_t(t))**moment
        if depth <= 0:
            return None

        At = self._A_x(x, s=s, t=t, moment=moment, b=endowment, endowment=0,
                       depth=depth-1)  #1-abs(depth))
        A = self._A_x(x, s=s, t=t, b=endowment, endowment=endowment, 
                      moment=moment, depth=1-abs(depth))        
        if A is not None and At is not None:
            self.blog(Blog.E(x=x, s=s, t=t), '=',
                      Blog.A(x=x, s=s, t=t, endowment=endowment), '-',
                      Blog.A(x=x, s=s, t=t, endowment=0),
                      #f"endowment - term insurance = {t}_E_{x+s}",
                      depth=depth, rule='endowment insurance minus term')
            return A - At  # (3) endowment insurance - term (helpful SULT)

        E = self._E_x(x, s=s, moment=moment, depth=1-abs(depth))
        Et = self._E_x(x, s=s+1, t=t-1, moment=moment, depth=depth-1)
        if E is not None and Et is not None:
            msg = f"chain Rule: {t}_E_{x+s} = E_{x+s} * {t-1}_E_{x+s+1}"
            self.blog(Blog.E(x=x, s=s, t=t, moment=moment), '=',
                      Blog.E(x=x, s=s, t=1, moment=moment), '*',
                      Blog.E(x=x, s=s+1, t=t-1, moment=moment), '*',
                      Blog.m(moment, endow=endowment),
                      depth=depth, rule='pure endowment chain rule')
            return E * Et * endowment**moment # (4) chain rule

    def E_x(self, x: int, s: int = 0, t: int = 1, 
            endowment: int = 1, moment: int = 1) -> float:
        """Compute pure endowment by calling recursion helper

        Args:
          x : age of selection
          s : years after selection
          t : term of pure endowment
          endowment : amount of pure endowment
          moment : compute first or second moment
        """
        self.blog = Blog("Pure Endowment",
                         Blog.E(x=x, s=s, t=t, moment=moment, endowment=endowment),
                         levels=self.maxdepth)
        if moment == self._VARIANCE:  # Bernoulli shortcut for variance
            found = self.get_E(x, s=s, t=t, endowment=endowment, moment=moment)
            if found is not None:
                return found
            t_p_x = self.p_x(x, s=s, t=t)
            return (endowment * self.interest.v_t(t))**2 * t_p_x * (1-t_p_x)
        found = self._E_x(x, s=s, t=t, endowment=endowment, moment=moment,
                          depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found

    #
    # Formulas for Increasing Insurance: IA_x:t
    #
    def get_IA(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
               b: int = 1, discrete: bool = True) -> float | None:
        """Get increasing insurance from key-value store

        Args:
          x : age of selection
          s : years after selection
          t : term of increasing insurance
          b : benefit after year 1
          discrete : discrete or continuous increasing insurance
        """
        key = self.db_key('IA', x=x+s, t=t, discrete=discrete)
        val = self.db.get(key, None)
        if val is not None:
            return val * b   # stored with benefit=1

    def set_IA(self, val: float, x: int, s: int = 0, t: int = Reserves._WHOLE,
               b: int = 1, discrete: bool = True) -> "Recursion":
        """Set increasing insurance IA_[x]+s:t to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          t : term of increasing insurance
          b : benefit after year 1
          discrete : discrete or continuous increasing insurance
        """
        val /= b   # store with benefit=1
        return self.db_put(self.db_key('IA', x=x+s, t=t, 
                                       discrete=discrete), val)

    def _IA_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, b: int = 1,
              discrete: bool = True, depth: int = 1) -> float | None:
        """Helper to compute from recursive and alternate formulas"""
        if t == 0:
            return 0
        found = self.get_IA(x=x, s=s, t=t, b=b, discrete=discrete)
        if found is not None:
            return found
        if depth <= 0:
            return None

        if t > 0:  # decreasing must be term insurance
            A = self._A_x(x=x, s=s, t=t, b=b, discrete=discrete, depth=depth-1)
            n = t + int(discrete)
            DA = self._DA_x(x=x, s=s, t=t, b=b, discrete=discrete, depth=depth-1)
            if A is not None and DA is not None:
                self.blog(Blog.IA(x=x, s=s, t=t), f'= {n}',
                          Blog.A(x=x, s=s, t=t), '-', Blog.DA(x=x, s=s, t=t),
                    #f"identity IA_{x+s}:{t}: ({n})A - DA",
                          depth=depth, rule='varying insurance identity')
                return A * n - DA  # (1) Identity with term and decreasing

        A = self._A_x(x=x, s=s, t=1, b=b, discrete=discrete, depth=depth-1)
        IA = self._IA_x(x=x, s=s+1, t=self.add_term(t, -1), b=b, depth=depth-1)
        p = self._p_x(x, s=s, t=1, depth=1-abs(depth))   # FIXED t=1
        if A is not None and IA is not None and p is not None:
            self.blog(Blog.IA(x=x, s=s, t=t), '=', Blog.A(x=x, s=s, t=t), '+',
                      Blog.p(x=x, s=s), f"* v(t={t}) *", Blog.IA(x=x, s=s+1, t=t-1),
                      #f"backward IA_{x+s}:{t}: A + IA_{x+s+1}:{t-1}",
                      depth=depth, rule='backward recursion')
            return A + p * self.interest.v * IA  # (2) backward recursion

    def increasing_insurance(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
                             b: int = 1, discrete: bool = True) -> float:
        """Compute increasing insurance with recursive helper

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit in first year
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        self.blog = Blog("Increasing Insurance",
                         Blog.IA(x=x, s=s, t=t, b=b, discrete=discrete),
                         levels=self.maxdepth)
        IA = self._IA_x(x, s=s, b=b, t=t, discrete=discrete, depth=self.maxdepth)
        if IA is not None:
            print(self.blog, end='')
            return IA
        IA = super().increasing_insurance(x, s=s, b=b, t=t, discrete=discrete)
        if IA is not None:
            print(self.blog, end='')
            return IA

    #
    # Formulas for Decreasing insurance: DA_x:t
    #
    def get_DA(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
               b: int = 1, discrete: bool = True) -> float | None:
        """Get decreasing insurance from key-value store

        Args:
          x : age of selection
          s : years after selection
          t : term of decreasing insurance
          b : benefit after year 1
          discrete : discrete or continuous decreasing insurance
        """
        key = self.db_key('DA', x=x+s, t=t, discrete=discrete)
        val = self.db.get(key, None)
        if val is not None:
            return val * b   # stored with benefit=1

    def set_DA(self, val: float, x: int, s: int = 0, t: int = Reserves._WHOLE,
               b: int = 1, discrete: bool = True) -> "Recursion":
        """Set decreasing insurance DA_[x]+s:t to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          t : term of decreasing insurance
          b : benefit after year 1
          discrete : discrete or continuous decreasing insurance
        """
        val /= b     # store with benefit=1
        return self.db_put(self.db_key('DA', x=x+s, t=t, 
                                       discrete=discrete), val)

    def _DA_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, b: int = 1,
              discrete: bool = True, depth: int = 1) -> float | None:
        """Helper to compute from recursive and alternate formulas"""
        found = self.get_DA(x=x, s=s, t=t, discrete=discrete)
        if found is not None:
            return found
        if t == 0:
            return 0
        if depth <= 0:
            return None
        if t < 0:  # decreasing must be term insurance
            return None
    
        A = self._A_x(x=x, s=s, t=t, b=b, discrete=discrete, depth=depth-1)
        n = t + int(discrete)
        IA = self._DA_x(x=x, s=s, t=t, b=b, discrete=discrete, depth=depth-1)
        if A is not None and IA is not None:
            self.blog(Blog.DA(x=x, s=s, t=t), f'= {n}',
                      Blog.A(x=x, s=s, t=t), '-', Blog.IA(x=x, s=s, t=t),
                      depth=depth, rule='varying insurance identity')
            return A * n - IA  # (1) identity with term and decreasing

        DA = self._IA_x(x=x, s=s+1, t=self.add_term(t, -1), b=b, depth=depth-1)
        p = self._p_x(x, s=s, depth=1-abs(depth))
        if DA is not None and p is not None:
            #msg = f"backward DA_{x+s}:{t}: v(t q_{x+s} + p_{x+s} DA_{x+s+1}:{t-1})"
            self.blog(Blog.DA(x=x, s=s, t=t), f"= v(t={t}) * t *", Blog.q(x=x, s=s),
                      '+', Blog.p(x=x, s=s), '*', Blog.DA(x=x, s=s+1, t=t-1),
                      depth=depth, rule='backward recursion')
            return self.interest.v * ((1-p)*t + p*DA)  # (2) backward recursion

    def decreasing_insurance(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
                             b: int = 1, discrete: bool = True) -> float:
        """Compute decreasing insurance by calling recursive helper first

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit in first year
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        self.blog = Blog("Increasing Insurance",
                         Blog.DA(x=x, s=s, t=t, b=b, discrete=discrete),
                         levels=self.maxdepth)
        if t == 0:
            return 0
        A = self._DA_x(x=x, s=s, t=t, b=b, discrete=discrete, depth=self.maxdepth)
        if A is not None:
            print(self.blog, end='')
            return A
        A = super().decreasing_insurance(x, s=s, b=b, t=t, discrete=discrete)
        if A is not None:
            print(self.blog, end='')
            return A

    #
    # Formulas for Insurance: A_x:t
    #
    def get_A(self, x: int, s: int = 0, u: int = 0, t: int = Reserves._WHOLE,
              b: int = 1, moment: int = 1, endowment: int = 0, 
              discrete: bool = True) -> float | None:
        """Get insurance from key-value store

        Args:
          x : age of selection
          s : years after selection
          u : defer u years
          t : term of insurance
          endowment : endowment amount
          discrete : discrete (True) or continuous (False) insurance
          moment : first or second moment of insurance
        """
        if endowment < 0:   # endowment insurance with equal benefits
            endowment = b
        if t == 0:          # terminal value of insurance
            return endowment
        if b == 0:          # normalize insurance factor by benefit amount
            scale = endowment
            endowment = 1
        else:
            scale = b
            endowment = 1 if b == endowment else endowment / b
        key = self.db_key('A', x=x+s, u=u, t=t, moment=moment, 
                          endowment=endowment, discrete=discrete)
        val = self.db.get(key, None)
        if val is not None:
            return val * scale   # stored with benefit=1


    def set_A(self, val: float, x: int, s: int = 0, t: int = Reserves._WHOLE,
              u: int = 0, b: int = 1, moment: int = 1, endowment: int = 0, 
              discrete: bool = True) -> "Recursion":
        """Set insurance u|_A_[x]+s:t to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          u : defer u years
          t : term of insurance
          endowment : endowment amount
          discrete : discrete (True) or continuous (False) insurance
          moment : first or second moment of insurance
        """
        if endowment < 0:      # endowment insurance with equal death and endow
            endowment = b
        if endowment == 0:     # normalize insurance factor by benefit amount
            val /= b
        elif b == 0:
            val /= endowment   # store with benefit=1
        else:
            if b != 1:
                val /= b
                endowment /= b
        return self.db_put(self.db_key('A', x=x+s, t=t, u=u,
                                       moment=moment, endowment=endowment, 
                                       discrete=discrete), val)

    def _A_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, u: int = 0,
             b: int = 1, discrete: bool = True, endowment: int = 0,
             moment: int = 1, depth: int = 1) -> float | None:
        """Helper to compute from recursive and alternate formulas"""
        
        if endowment == b and t == 1 and discrete: # 1-year endow ins
            return (self.interest.v_t(1) * endowment)**moment
        found = self.get_A(x=x, s=s, t=t, b=b, u=u, discrete=discrete, 
                           moment=moment, endowment=endowment)
        if found is not None:
            return found
        if depth <= 0:
            return None

        if u > 0:  # (1) deferred insurance  
            A = self._A_x(x=x, s=s+1, t=t, b=b, u=u-1, discrete=discrete, 
                              moment=moment, endowment=endowment, depth=depth-1)
            E = self._E_x(x, s=s, t=1, moment=moment, depth=depth-1)
            if A is not None and p is not None:  # (1a) backward E_x * A
                #msg = f"backward deferred {u}_A_{x+s}: {u}_E * A_{x+s+u}"
                self.blog(Blog.A(x=x, s=s, t=t, u=u, b=b, moment=moment), '=',
                          Blog.E(x=x, s=s, moment=moment), '*',
                          Blog.A(x=x, s=s+1, t=t, u=u-1, moment=moment),
                          depth=depth, rule='backward recursion')
                return E * A

            A = self._A_x(x, s=s-1, t=t, b=b, u=u+1, discrete=discrete, 
                          moment=moment, endowment=endowment, depth=depth-1)
            E = self._E_x(x, s=s-1, t=1, moment=moment, depth=depth-1)
            if A is not None and E is not None: # (1b) forward recursion
                msg = f"forward deferred {u}_A_{x+s}: {u+1}A_{x+s-1} / E"
                self.blog(Blog.A(x=x, s=s, t=t, u=u, b=b, moment=moment), '=',
                          Blog.A(x=x, s=s-1, t=t, u=u+1, moment=moment), '/',
                          Blog.E(x=x, s=s-1, moment=moment),
                          depth=depth, rule='forward recursion')
                return A / E
            return None

        if endowment > 0: # (2a) endowment = term + E_x * endowment
            A = self._A_x(x=x, s=s, t=t, b=b, discrete=discrete, 
                          moment=moment, depth=1-abs(depth))
            E_x = self._E_x(x=x, s=s, t=t, moment=moment, 
                            endowment=endowment, depth=1-abs(depth))
            if A is not None and E_x is not None:
                # f"term + pure insurance = A_{x+s}:{t}",
                self.blog(Blog.A(x=x, s=s, t=t, endowment=endowment, moment=moment),
                          '=', Blog.A(x=x, s=s, t=t, moment=moment), '+',
                          Blog.E(x=x, s=s, t=t, endowment=endowment, moment=moment),
                          depth=depth, rule='term plus pure endowment')
                return A + E_x
        else:             # (2b) term = endowment insurance - E_x * endowment
            A = self._A_x(x=x, s=s, t=t, b=b, discrete=discrete, 
                          moment=moment, endowment=b, depth=1-abs(depth))
            E_x = self._E_x(x=x, s=s, t=t, moment=moment, endowment=b, 
                            depth=1-abs(depth))
            if A is not None and E_x is not None:
                #msg = f"endowment insurance - pure endowment = A_{x+s}^1:{t}"
                self.blog(Blog.A(x=x, s=s, t=t, moment=moment), '=',
                          Blog.A(x=x, s=s, t=t, moment=moment, endowment=b), '-',
                          Blog.E(x=x, s=s, t=t, endowment=b, moment=moment),
                          depth=depth, rule='endowment insurance - pure')
                return A - E_x

        if not discrete:  # recursions for discrete insurance
            return None
        if t == 1:  # special cases for discrete one-year insurance
            if endowment == b:  # (3a) discrete one-year endowment insurance
                self.blog(Blog.A(x=x, s=s, t=1, endowment=endowment,
                                 moment=moment), f"= v"+Blog.m(moment),
                          "*", Blog.m(moment, endow=endowment),
                          depth=depth, rule='one-year endowment insurance')
                return (self.interest.v * endowment)**moment
            p = self._p_x(x, s=s, t=1)
            if p is not None:  # (3b) one-year discrete insurance
                self.blog(Blog.A(x=x, s=s, t=t, moment=moment, endowment=endowment),
                          f"= v"+Blog.m(moment), "*",
                          Blog.q(x=x, s=s), f"*", "v"*Blog.m(moment), "+",
                          Blog.p(x=x, s=s), f"*".
                          Blog.m(moment, endow=endowment),
                          #f"discrete 1-year insurance: A_{x+s}:1 = qv",
                          depth=depth, rule='one-year discrete insurance')
                return (self.interest.v**moment 
                        * ((1 - p) * b**moment + p * endowment**moment))
    
        A = self._A_x(x=x, s=s+1, t=self.add_term(t, -1), b=b, 
                      discrete=discrete, moment=moment,
                      endowment=endowment, depth=depth-1)
        p = self._p_x(x, s=s, t=1, depth=1-abs(depth)) # (4) backward recursion
        if A is not None and p is not None:
            self.blog(Blog.A(x=x, s=s, t=t, b=b, moment=moment),
                      f"= v"+Blog.m(moment), "* [",
                      Blog.q(x=x,s=s), f"* b"+Blog.m(moment), "+",
                      Blog.p(x=x, s=s), '*',
                      Blog.A(x=x, s=s+1, t=t-1, b=b, moment=moment), ']',
                      #f"backward: A_{x+s} = qv + pvA_{x+s+1}",
                      depth=depth, rule='backward recursion')
            return self.interest.v_t(1)**moment * ((1 - p)*b**moment + p*A)

        A = self._A_x(x=x, s=s-1, t=self.add_term(t, 1), b=b, u=u, 
                      discrete=discrete, moment=moment, 
                      endowment=endowment, depth=depth-1)
        p = self._p_x(x, s=s-1, t=1, depth=1-abs(depth))
        if A is not None and p is not None:  # (5) forward recursion
            self.blog(Blog.A(x=x, s=s, t=t, b=b, moment=moment), '= [',
                      Blog.A(x=x, s=s-1, t=t+1, b=b, moment=moment),
                      f"/v"+Blog.m(moment), "-", Blog.q(x=x,s=s),
                      f"* b"+Blog.m(moment), "] /", Blog.p(x=x, s=s),
                      #f"forward: A_{x+s} = (A_{x+s-1}/v - q) / p",
                      depth=depth, rule='forward recursion')
            return (A/self.interest.v_t(1)**moment - (1-p)*b**moment) / p

    def whole_life_insurance(self, x: int, s: int = 0, b: int = 1, 
                             discrete: bool = True, moment: int = 1) -> float:
        """Compute whole life insurance A_x by calling recursion helper and twin

        Args:
          x : age of selection
          s : years after selection
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        self.blog = Blog("Whole Life Insurance",
                         Blog.A(x=x, s=s, t=Reserves._WHOLE, b=b, u=0,
                                endowment=0, moment=moment, discrete=discrete),
                         levels=self.maxdepth)
        found = self._A_x(x, s=s, b=b, moment=moment, discrete=discrete,
                          depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found
        if moment == 1 and self.interest.i > 0:  # (1) twin annuity
            a = self._a_x(x, s=s, b=b, discrete=discrete, depth=self.maxdepth)
            if a is not None:
                self.blog("A_x = 1 - d * a_x",
                          depth=self.maxdepth, rule='annuity twin')
                print(self.blog, end='')
                return self.insurance_twin(a=a, discrete=discrete)
        A = super().whole_life_insurance(x, s=s, b=b, discrete=discrete,
                                         moment=moment)
        if A is not None:
            print(self.blog, end='')
            return A

    def term_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1, 
                       moment: int = 1, discrete: bool = True) -> float:
        """Compute term life insurance A_x:t^1 by calling recursion helper: 

        Args:
          x : age of selection
          s : years after selection
          t : term of insurance
          b : amount of benefit
          moment : compute first or second moment
          discrete : benefit paid year-end (True) or moment of death (False)
        """
        self.blog = Blog("Term Insurance",
                         Blog.A(x=x, s=s, t=t, b=b, u=0, discrete=discrete,
                                endowment=0, moment=moment),
                         levels=self.maxdepth)
        found = self._A_x(x, s=s, b=b, t=t, moment=moment, discrete=discrete,
                          depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found
        A = super().term_insurance(x, s=s, b=b, t=t, discrete=discrete,
                                   moment=moment)
        if A is not None:
            print(self.blog, end='')
            return A

    def deferred_insurance(self, x: int, s: int = 0, b: int = 1, u: int = 0, 
                           t: int = Reserves._WHOLE, moment: int = 1, 
                           discrete: bool = True) -> float:
        """Compute deferred life insurance u|A_x:t^1 by calling recursion helper: """
        self.blog = Blog("Deferred Insurance",
                         Blog.A(x=x, s=s, t=t, b=b, u=u, discrete=discrete,
                                endowment=0, moment=moment),
                         levels=self.maxdepth)
        A = self.get_A(x=x, s=s, t=t, b=b, u=u, discrete=discrete, 
                       moment=moment)
        if A is not None:
            print(self.blog, end='')
            return A
        A = super().deferred_insurance(x, s=s, b=b, t=t, u=u, 
                                       discrete=discrete, moment=moment)
        if A is not None:
            print(self.blog, end='')
            return A
    
    def endowment_insurance(self, x: int, s: int = 0, t: int = 1, b: int = 1,
                            endowment: int = -1, moment: int = 1,
                            discrete: bool = True) -> float:
        """Compute endowment insurance u|A_x:t by calling recursion helper: """
        self.blog = Blog("Endowment Insurance",
                         Blog.A(x=x, s=s, t=t, b=b, u=0, discrete=discrete,
                                endowment=endowment, moment=moment),
                         levels=self.maxdepth)
        assert t >= 0
        if endowment < 0:
            endowment = b
        found = self._A_x(x, s=s, b=b, t=t, moment=moment, discrete=discrete,
                          endowment=endowment, depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found
        if moment == 1 and endowment == b and self.interest.i > 0:
            a = self._a_x(x, s=s, b=b, t=t, moment=moment, discrete=discrete,
                          depth=self.maxdepth)
            if a is not None:   # twin insurance
                print(self.blog, end='')
                return self.insurance_twin(a=a, discrete=discrete)
        A = super().endowment_insurance(x, s=s, b=b, t=t, discrete=discrete,
                                        moment=moment, endowment=endowment)
        if A is not None:
            print(self.blog, end='')
            return A

    #
    # Formulas for Annuties: a_x:t
    #
    def get_a(self, x: int, s: int = 0, u: int = 0, t: int = Reserves._WHOLE,
              b: int = 1, variance: bool = False, 
              discrete: bool = True) -> float | None:
        """Get annuity from key-value store

        Args:
          x : age of selection
          s : years after selection
          u : defer u years
          t : term of annuity
          b : benefit amount
          discrete : whether annuity due (True) or continuous (False)
          variance : whether first moment (False) or variance (True)
        """
        key = self.db_key('a', x=x+s, u=u, t=t, 
                          variance=variance, discrete=discrete)
        val = self.db.get(key, None)
        if val is not None:
            return val * b    # stored with benefit=1

    def set_a(self, val: float, x: int, s: int = 0, t: int = Reserves._WHOLE,
              u: int = 0, b: int = 1, variance: bool = False, 
              discrete: bool = True) -> "Recursion":
        """Set annuity u|_a_[x]+s:t to given value

        Args:
          val : value to set
          x : age of selection
          s : years after selection
          u : defer u years
          t : term of annuity
          b : benefit amount
          discrete : whether annuity due (True) or continuous (False)
          variance : whether first moment (False) or variance (True)
        """
        val /= b    # store with benefit=1
        return self.db_put(self.db_key('a', x=x+s, t=t, u=u, 
                           variance=variance, discrete=discrete), val)

    def _a_x(self, x: int, s: int = 0, t: int = Reserves._WHOLE, 
             u: int = 0, b: int = 1, discrete: bool = True, 
             variance: bool = False, depth: int = 1) -> float | None:
        """Helper to compute from recursive and alternate formulas"""
        if t == 1 and not u and discrete:
            self.blog(Blog.a(x=x, s=s, t=1, discrete=discrete), '= 1',
                      depth=depth, rule='one-year discrete annuity')
            return b 
        if t == 0:
            return 0
        found = self.get_a(x=x, s=s, t=t, b=b, u=u, discrete=discrete,
                           variance=variance)
        if found is not None:
            return found
        if depth <= 0:
            return None
        if variance:
            return None

        if u > 0:  # (1) deferred annuity
            found = self._a_x(x=x, s=s+1, t=t, b=b, u=u-1, discrete=discrete, 
                              variance=variance, depth=depth-1)
            E = self._E_x(x, s=s, t=1, depth=depth-1)
            if found is not None and E is not None:
                #msg = f"backward {u}_a_{x+s} = {u}_E * a_{x+s+u}"
                self.blog(Blog.a(x=x, s=s, u=u, t=t), '=',
                          Blog.a(x=x, s=s+1, t=t, u=u-1), '/', Blog.E(x=x, s=s),
                          depth=depth, rule='backward deferred annuity')
                return E * found  # (1a) backward recusion

            found = self._a_x(x=x, s=s-1, t=t, b=b, u=u+1, discrete=discrete, 
                              depth=depth-1)         # FIXED u=u+1
            E = self._E_x(x, s=s-1, t=1, depth=depth-1)
            if found is not None and E is not None:  # (1b) forward
                #msg = f"forward: {u}_a_{x+s} = {u+1}_a_{x+s-1}/E_{x+s-1}"
                self.blog(Blog.a(x=x, s=s, u=u, t=t), '=',
                          Blog.a(x=x, s=s-1, t=t, u=u+1), '/', Blog.E(x=x, s=s-1),
                          depth=depth, rule='forward deferred annuity')
                return found / E

        else:  # (2) Temporary and whole annuity
            found = self._a_x(x=x, s=s+1, t=self.add_term(t, -1), b=b, u=u, 
                              discrete=discrete, 
                              variance=variance, depth=depth-1)
            E = self._E_x(x, s=s, t=1, depth=depth-1)
            if found is not None and E is not None:  # (2a) backward
                #msg = (f"backward: a_{x+s}{'' if t < 0 else (':'+str(t))} = 1 + "
                #       + f"E_{x+s} a_{x+s+1}{'' if t < 0 else (':'+str(t-1))}")
                #_t = "" if t < 0 else f":{t-1}"
                self.blog(Blog.a(x=x, s=s, t=t), '= 1 +',
                          Blog.E(x=x, s=s, t=1), '*', Blog.a(x=x, s=s+1, t=t-1),
                          depth=depth, rule='backward recursion')
                return b + E * found

            found = self._a_x(x=x, s=s-1, t=self.add_term(t, 1), b=b, u=u, 
                              discrete=discrete, depth=depth-1)
            E = self._E_x(x, s=s-1, t=1, depth=depth-1)
            if found is not None and E is not None:  # (2b) forward
                _t = "" if t < 0 else f":{t-1}"
                self.blog(Blog.a(x=x, s=s, t=t), '= [',
                          Blog.a(x=x, s=s-1), '- 1 ] /',
                          Blog.E(x=x, s=s-1, t=1),
                    #f"forward: a_{x+s}{_t} = (a_{x+s-1} - 1)/E",
                          depth=depth, rule='forward recursion')
                return (found - b) / E

    def whole_life_annuity(self, x: int, s: int = 0, b: int = 1, 
                           variance: bool = False,
                           discrete: bool = True) -> float:
        """Compute whole life annuity a_x by calling recursion then twin first

        Args:
          x : age of selection
          s : years after selection
          b : annuity benefit amount
          variance (bool): return EPV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        self.blog = Blog("Whole Life Annuity",
                         Blog.a(x=x, s=s, t=Reserves._WHOLE, b=b, u=0,
                                discrete=discrete, variance=False),
                         levels=self.maxdepth)
        found = self._a_x(x, s=s, b=b, variance=variance, 
                          discrete=discrete, depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found
        if not variance and self.interest.i > 0:  # (1) twin insurance shortcut
            A = self._A_x(x, s=s, b=b, discrete=discrete, depth=self.maxdepth)
            if A is not None:
                self.blog("a_x = (1-A_x) / d",
                          depth=self.maxdepth, rule='insurance twin')
                print(self.blog, end='')
                return self.annuity_twin(A=A, discrete=discrete)
        a = super().whole_life_annuity(x, s=s, b=b, discrete=discrete,
                                       variance=variance)
        if a is not None:
            print(self.blog, end='') 
            return a

    def temporary_annuity(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
                          b: int = 1, variance: bool = False,
                          discrete: bool = True) -> float:
        """Compute temporary annuity a_x:t by calling recursion then twin first

        Args:
          x : age of selection
          s : years after selection
          t : term of annuity in years
          b : annuity benefit amount
          variance (bool): return EPV (True) or variance (False)
          discrete : annuity due (True) or continuous (False)
        """
        self.blog = Blog("Temporary Annuity",
                         Blog.a(x=x, s=s, t=t, b=b, u=0,
                                discrete=discrete, variance=variance),
                         levels=self.maxdepth)
        found = self._a_x(x, s=s, b=b, t=t, variance=variance, 
                          discrete=discrete, depth=self.maxdepth)
        if found is not None:
            print(self.blog, end='')
            return found
        if not variance and self.interest.i > 0: # (1) twin insurance shortcut
            A = self._A_x(x, s=s, b=b, t=t, endowment=b, discrete=discrete, 
                          depth=self.maxdepth)
            if A is not None:
                self.blog(Blog.a(x=x, s=s, t=t), '= [ 1 -',
                          Blog.A(x=x, s=s, t=t), f"] / d(t={t})",
                          #"Annuity twin: a = (1 - A) / d",
                          depth=self.maxdepth, rule='annuity twin')
                print(self.blog, end='')
                return self.annuity_twin(A=A, discrete=discrete)
        a = super().temporary_annuity(x, s=s, b=b, t=t, discrete=discrete,
                                      variance=variance)
        if a is not None:
            print(self.blog, end='')
            return a

    def deferred_annuity(self, x: int, s: int = 0, t: int = Reserves._WHOLE,
                         u: int = 0, b: int = 1, discrete: bool = True) -> float:
        """Compute deferred annuity u|a_x:t by calling recursion first

        Args:
          x : age of selection
          s : years after selection
          u : years deferred
          t : term of annuity in years
          b : annuity benefit amount
          discrete : annuity due (True) or continuous (False)
        """
        self.blog = Blog("Deferred Annuity",
                         Blog.a(x=x, s=s, t=t, b=b, u=u,
                                discrete=discrete, variance=False),
                         levels=self.maxdepth)
        a = self._a_x(x, s=s, b=b, t=t, u=u, 
                          discrete=discrete, depth=self.maxdepth)
        if a is not None:
            print(self.blog, end='')
            return a
        a = self._a_x(x, s=s, b=b, t=self.add_term(u, t),  
                      discrete=discrete, depth=self.maxdepth)
        a_t = self._a_x(x, s=s, b=b, t=u,  
                        discrete=discrete, depth=self.maxdepth)
        if a is not None and a_t is not None:
            print(self.blog, end='')
            return a - a_t
        return super().deferred_annuity(x, s=s, b=b, t=t, discrete=discrete)


class Blog:
    """Helper to track and display recursion progress"""
    _verbose : bool = True
    def __init__(self, *args, levels: int = 3, width: int = 80):
        self.title = ' *' + " ".join(args) + " <--"  # to identify this stack
        self.levels = levels                         # maximum levels to indent
        self.width = width - 3                       # display line width
        self._history = []
        self._rules = []
        self._depths = []

    def __call__(self, *args, depth: int | None = None, rule: str = ''):
        """Append next message to history"""
        assert depth is not None and rule, "depth and rule must be specified"
        msg = " ".join([a for a in args]) 
        if msg not in self._history:
            self._history.insert(0, msg)
            self._depths.insert(0, depth)
            self._rules.insert(0, rule)

    def __len__(self) -> int:
        return int(Blog._verbose and bool(self._history))

    def __str__(self):
        """Display message history"""
        if not len(self):
            return ''
        _str = self.title + '\n'
        for msg, depth, rule in zip(self._history, self._depths, self._rules):
            left = ' '*(3+max(self.levels-abs(depth), 0))
            right = ' '*max(5, self.width - len(msg) - len(left) - len(rule))
            _str += left + msg + right + '~' + rule + '\n'
        return _str

    @staticmethod
    def q(x: int, s: int = 0, t: int = 1, u: int = 0) -> str:
        """Return string representation of mortality u|t_q_x term"""
        out = []
        if t != 1:
            out.append(f"t={t}")
        if u > 0:
            out.append(f"defer={u}")
        return f"q_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def p(x: int, s: int = 0, t: int = 1) -> str:
        """Return string representation of survival t_p_x term"""
        out = []
        if t != 1:
            out.append(f"t={t}")
        return f"p_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)


    @staticmethod
    def e(x: int, s: int = 0, t: int = Reserves._WHOLE, curtate: bool = False,
          moment: int = 1) -> str:
        """Return string representation of expected future lifetime e_x term"""
        out = []
        if t >= 0:
            out.append(f"t={t}")
        if moment != 1:
            out.append(f"mom={moment}")
        out.append('curtate' if curtate else 'complete')
        return f"e_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def E(x: int, s: int = 0, t: int = 1, endowment: int = 1,
          moment: int = 1) -> str:
        """Return string representation of pure endowment t_E_x term"""
        out = []
        out.append(f"t={t if t >= 0 else 'WL'}")  # term or whole life
        if moment != 1:
            out.append(f"mom={moment}")
        if endowment != 1:
            out.append(f"endow={endowment}")
        return f"E_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)
    
    @staticmethod
    def IA(x: int, s: int = 0, t: int = Reserves._WHOLE, b: int = 1,
           discrete: bool = True) -> str:
        """Return string representation of increasing insurance IA_x term"""
        out = []
        out.append(f"t={t if t >= 0 else 'WL'}")  # term or whole life
        if b != 1:
            out.append(f"b={b}")
        return f"IA_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def DA(x: int, s: int = 0, t: int = Reserves._WHOLE, b: int = 1,
           discrete: bool = True) -> str:
        """Return string representation of decreasing insurance DA_x term"""
        out = []
        out.append(f"t={t if t >= 0 else 'WL'}")  # term or whole life
        if b != 1:
            out.append(f"b={b}")
        return f"DA_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def A(x: int, s: int = 0, t: int = Reserves.WHOLE, u: int = 0, b: int = 1,
          moment: int = 1, endowment: int = 0, discrete: bool = True) -> str:
        """Return string representation of insurance A_x term"""
        out = []
        out.append(f"t={t if t >= 0 else 'WL'}")  # term or whole life
        if u != 0:
            out.append(f"u={u}")
        if b != 1:
            out.append(f"b={b}")
        if endowment != 0:
            out.append(f"endow={endowment}")
        if moment != 1:
            out.append(f"mom={moment}")
        return f"A_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def a(x: int, s: int = 0, t: int = Reserves.WHOLE, u: int = 0, b: int = 1,
          variance: bool = False, discrete: bool = True) -> str:
        """Return string representation of annuity a_x term"""
        out = []
        out.append(f"t={t if t >= 0 else 'WL'}")  # term or whole life
        if u != 0:
            out.append(f"u={u}")
        if b != 1:
            out.append(f"b={b}")
        if variance:
            out.append('var')
        return f"a_{x+s}" + "("*bool(out) + ",".join(out) + ")"*bool(out)

    @staticmethod
    def m(moment: int, **kwargs) -> str:
        """Return string representation of moment exponent"""
        out = f"^{moment}"*(moment != 1)
        if not kwargs:
            return out
        args = [k for k,v in kwargs.items() if v]
        return ("(" + "*".join(args) + ")")+out if args else "0"
    

if __name__ == "__main__":
    from actuarialmath.constantforce import ConstantForce
    from actuarialmath.policyvalues import Contract

    print("SOA Question 6.10:  ")
    x = 0
    life = Recursion(verbose=True, depth=5).set_interest(i=0.06).set_p(0.975, x=x)
    a = 152.85/56.05  # solve a_x:3, given net premium and benefit APV
    life.set_a(a, x=x, t=3)
    #life.set_A(152.85, x=x, t=3, b=1000)

    p1 = life.p_x(x=x+1)   # buggy
#    p2 = life.p_x(x=x+2)
    life.set_p(p1, x=x+1)

    def fun(p): 
        return life.set_p(p, x=x+2).term_insurance(x=x, t=3, b=1000)
    p = life.solve(fun, target=152.85, grid=0.975)  # solve p_x+2, given A_x:3
    print(p)

    #isclose(0.91, p, question="Q6.10")

    
    print("AMLCR2 Exercise 2.6")
    x = 0
    life = Recursion(depth=3).set_interest(i=0.06)\
                             .set_p(0.99, x=x)\
                             .set_p(0.985, x=x+1)\
                             .set_p(0.95, x=x+1, t=3)\
                             .set_q(0.02, x=x+3)

    print(life.p_x(x=x+3))  # 0.98
    print(life.p_x(x=x, t=2))  # 0.97515
    print(life.p_x(x=x+1, t=2))  # 0.96939
    print(life.p_x(x=x, t=3))  # 0.95969
    print(life.q_x(x=x, t=2, u=1))  # 0.03031

    print("SOA Question 6.48:  (A) 3195")
    life = Recursion().set_interest(i=0.06)
    x = 0
    life.set_p(0.95, x=x, t=5)
    life.set_q(0.02, x=x+5)
    life.set_q(0.03, x=x+6)
    life.set_q(0.04, x=x+7)
    a = 1 + life.E_x(x, t=5)
    A = life.deferred_insurance(x, u=5, t=3)
    P = life.gross_premium(A=A, a=a, benefit=100000)
    print(P)
    print()


    print("SOA Question 6.40: (C) 116 ")
    # - standard formula discounts/accumulates by too much (i should be smaller)
    x = 0
    life = Recursion().set_interest(i=0.06).set_a(7, x=x+1).set_q(0.05, x=x)
    a = life.whole_life_annuity(x)
    A = 110 * a / 1000
    print(a, A)
    life = Recursion().set_interest(i=0.06).set_A(A, x=x).set_q(0.05, x=x)
    A1 = life.whole_life_insurance(x+1)
    P = life.gross_premium(A=A1 / 1.03, a=7) * 1000
    print(P)
    print()
    
    print("SOA Question 6.17:  (A) -30000")
    x = 0
    life = ConstantForce(mu=0.1).set_interest(i=0.08)
    A = life.endowment_insurance(x, t=2, b=100000, endowment=30000)
    a = life.temporary_annuity(x, t=2)
    P = life.gross_premium(a=a, A=A)
    print(A, a, P)

    life1 = Recursion().set_interest(i=0.08)\
                       .set_q(life.q_x(x, t=1) * 1.5, x=x, t=1)\
                       .set_q(life.q_x(x+1, t=1) * 1.5, x=x+1, t=1)
    contract = Contract(premium=P * 2, benefit=100000, endowment=30000)
    L = life1.gross_policy_value(x, t=0, n=2, contract=contract)
    print(L)
    print()

    life = Recursion(verbose=False).set_interest(i=0.05).set_E(0.95, 0,t=1)
    E = life.E_x(0, t=1, moment=2)
    print(E)



"""Life Contingent Risks - Applies probability laws

MIT License. Copyright (c) 2022-2023 Terence Lim
"""
import math
import numpy as np
import pandas as pd
from scipy.special import ndtri
from scipy.stats import norm
from typing import Callable, Dict, Any, Tuple, List
from actuarialmath.interest import Interest
from actuarialmath.actuarial import Actuarial

class Life(Actuarial):
    """Compute moments and probabilities"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_interest(i=0)

    def set_interest(self, **interest) -> "Life":
        """Set interest rate, which can be given in any form

        Args:
          i : assumed annual interest rate
          d : or assumed discount rate
          v : or assumed discount factor
          delta : or assumed contiuously compounded interest rate
          v_t : or assumed discount rate as a function of time
          i_m : or assumed monthly interest rate
          d_m : or assumed monthly discount rate
          m : m'thly frequency, if i_m or d_m are given
        """
        self.interest = Interest(**interest)
        return self

    #
    # Probability theory
    #
    @staticmethod
    def variance(a, b, var_a, var_b, cov_ab: float) -> float:
        """Variance of weighted sum of two r.v.

        Args:
          a : weight on first r.v.
          b : weight on other r.v.
          var_a : variance of first r.v.
          var_b : variance of other r.v.
          cov_ab : covariance of the r.v.'s
        """
        return a**2 * var_a + b**2 * var_b + 2 * a * b * cov_ab

    @staticmethod
    def covariance(a, b, ab: float) -> float:
        """Covariance of two r.v.

        Args:
          a : expected value of first r.v.
          b : expected value of other r.v.
          ab : expected value of product of the two r.v.
        """
        return ab - a * b  # Cov(X,Y) = E[XY] - E[X] E[Y]

    @staticmethod
    def bernoulli(p, a: float = 1, b: float = 0,
                  variance: bool = False) -> float:
        """Mean or variance of bernoulli r.v. with values {a, b}

        Args:
          p : probability of first value
          a : first value
          b : other value
          variance : whether to return variance (True) or mean (False)
        """
        assert 0 <= p <= 1.
        return (a - b)**2 * p * (1-p) if variance else p * a + (1-p) * b

    @staticmethod
    def binomial(p: float, N: int, variance: bool = False) -> float:
        """Mean or variance of binomial r.v.

        Args:
          p : probability of occurence
          N : number of trials
          variance : whether to return variance (True) or mean (False)
        """
        assert 0 <= p <= 1. and N >= 1
        return N * p * (1-p) if variance else N * p

    @staticmethod
    def mixture(p, p1, p2: float, N: int = 1, variance: bool = False) -> float:
        """Mean or variance of binomial mixture

        Args:
          p : probability of selecting first r.v.
          p1 : probability of occurrence if first r.v.
          p2 : probability of occurrence if other r.v.
          N : number of trials
          variance : whether to return variance (True) or mean (False)
        """
        assert 0 <= p <= 1 and 0 <= p1 <= 1 and 0 <= p2 <= 1 and N >= 1
        mean1 = Life.binomial(p1, N)
        mean2 = Life.binomial(p2, N)
        if variance:
            var1 = Life.binomial(p1, N, variance=True)
            var2 = Life.binomial(p2, N, variance=True)
            return (Life.bernoulli(p, mean1**2 + var1, mean2**2 + var2) -
                    Life.bernoulli(p, mean1, mean2)**2)
        else:
            return Life.bernoulli(p, mean1, mean2)
        
    @staticmethod
    def conditional_variance(p, p1, p2: float, N: int = 1) -> float:
        """Conditional variance formula

        Args:
          p : probability of selecting first r.v.
          p1 : probability of occurence for first r.v.
          p2 : probability of occurence for other r.v.
          N : number of trials
        """
        assert 0 <= p <= 1 and 0 <= p1 <= 1 and 0 <= p2 <= 1 and N >= 1
        mean1 = Life.binomial(p1, N)
        mean2 = Life.binomial(p2, N)
        var1 = Life.binomial(p1, N, variance=True)
        var2 = Life.binomial(p2, N, variance=True)
        return (Life.bernoulli(p, mean1, mean2, variance=True)  # var of mean
                + Life.bernoulli(p, var1, var2))           # plus mean of var

    @staticmethod
    def portfolio_percentile(mean: float, variance: float,
                             prob: float, N: int = 1) -> float:
        """Probability percentile of the sum of N iid r.v.'s

        Args:
          mean : mean of each independent obsevation
          variance : variance of each independent observation
          prob : probability threshold
          N : number of observations to sum
        """
        assert prob < 1.0
        mean *= N
        variance *= N
        return mean + ndtri(prob) * math.sqrt(variance)

    @staticmethod
    def portfolio_cdf(mean: float, variance: float, value: float,
                      N: int = 1) -> float:
        """Probability distribution of a value in the sum of N iid r.v.

        Args:
          mean : mean of each independent obsevation
          variance : variance of each independent observation
          value : value to compute probability distribution in the sum
          N : number of observations to sum
        """
        mean *= N
        variance *= N
        return norm.cdf(value, loc=mean, scale=math.sqrt(variance))

    @staticmethod
    def quantiles_frame(quantiles: List[float] = [.8, .85, .9, .95,
                                                  .975, .99, .995]) -> Any:
        """Display selected quantile values from Normal distribution table

        Args:
          quantiles : list of quantiles to display normal distribution values
        """
        columns = [round(Life.portfolio_percentile(0, 1, p), 3) for p in quantiles]
        tab = pd.DataFrame.from_dict(data={'Pr(Z<=z)': quantiles}, 
                                     columns=columns, orient='index')\
                                    .rename_axis('z', axis="columns")
        return tab.round(3)


if __name__ == "__main__":
    print("SOA Question 2.2: (D) 400")
    p1 = (1. - 0.02) * (1. - 0.01)  # 2_p_x if vaccine given
    p2 = (1. - 0.02) * (1. - 0.02)  # 2_p_x if vaccine not given
    cond = math.sqrt(Life.conditional_variance(p=.2, p1=p1, p2=p2, N=100000))
    print(cond)   # conditional variance formula
    mix = math.sqrt(Life.mixture(p=.2, p1=p1, p2=p2, N=100000, variance=True))
    print(mix)    # mixture of distributions formula

    print()
    print("Values of z for selected values of Pr(Z<=z)")
    print("-------------------------------------------")
    print(Life.quantiles_frame().to_string(float_format=lambda x: f"{x:.3f}"))
    print()

